from typing import List, Any
from abc import ABC, abstractmethod
import gymnasium as gym
from csle_common.dao.simulation_config.simulation_trace import SimulationTrace


class BaseEnv(gym.Env, ABC):  # type: ignore
    """
    Abstract class representing a csle Environment
    """

    @abstractmethod
    def get_traces(self) -> List[SimulationTrace]:
        """
        :return: the list of simulation traces
        """
        pass

    @abstractmethod
    def reset_traces(self) -> None:
        """
        Resets the list of traces

        :return: None
        """
        pass

    @abstractmethod
    def manual_play(self) -> None:
        """
        An interactive loop for manual play of the environment

        :return: None
        """
        pass

    @abstractmethod
    def set_model(self, model: Any) -> None:
        """
        Sets the model. Useful when using RL frameworks where the stage policy is not easy to extract

        :param model: the model
        :return: None
        """
        pass

    @abstractmethod
    def set_state(self, state: Any) -> None:
        """
        Sets the state. Allows to simulate samples from specific states

        :param state: the state
        :return: None
        """
        pass



from unittest.mock import MagicMock
from csle_common.consumer_threads.aggregated_ossec_ids_log_consumer_thread import AggregatedOSSECIdsLogConsumerThread


class TestSuiteAggregatedOSSECIdsLogConsumerThread:
    """
    Test suite for AggregatedOSSECIdsLogConsumerThread
    """

    def test_initialization(self) -> None:
        """
        Initializes the test

        :return: None
        """
        kafka_server_ip = "127.0.0.1"
        kafka_port = 9092
        alert_counters = MagicMock()
        auto_offset_reset = "earliest"

        thread = AggregatedOSSECIdsLogConsumerThread(
            kafka_server_ip=kafka_server_ip,
            kafka_port=kafka_port,
            ossec_ids_alert_counters=alert_counters,
            auto_offset_reset=auto_offset_reset,
        )
        assert thread.running
        assert thread.kafka_server_ip == kafka_server_ip
        assert thread.kafka_port == kafka_port
        assert thread.ossec_ids_alert_counters == alert_counters

    def test_get_aggregated_ids_alert_counters(self) -> None:
        """
        Test the method that returns the aggregated alert counters from the list

        :return: None
        """
        mock_alert_counters_1 = MagicMock()
        mock_alert_counters_2 = MagicMock()
        mock_alert_counters_1.copy.return_value = mock_alert_counters_1
        mock_alert_counters_2.copy.return_value = mock_alert_counters_2
        mock_alert_counters_1.add = MagicMock()
        kafka_server_ip = "127.0.0.1"
        kafka_port = 9092
        alert_counters = MagicMock()
        auto_offset_reset = "earliest"
        thread = AggregatedOSSECIdsLogConsumerThread(
            kafka_server_ip=kafka_server_ip,
            kafka_port=kafka_port,
            ossec_ids_alert_counters=alert_counters,
            auto_offset_reset=auto_offset_reset,
        )
        thread.ossec_ids_alert_counters_list = []
        thread.ossec_ids_alert_counters = mock_alert_counters_1

        result = thread.get_aggregated_ids_alert_counters()
        assert result == mock_alert_counters_1.copy.return_value



import grpc


class GrpcUtil:
    """
    Class with utility functions for GRPC functionality in CSLE
    """

    @staticmethod
    def grpc_server_on(channel) -> bool:
        """
        Utility function to test if a given gRPC channel is working or not

        :param channel: the channel to test
        :return: True if working, False if timeout
        """
        try:
            grpc.channel_ready_future(channel).result(timeout=15)
            return True
        except grpc.FutureTimeoutError:
            return False



import logging
from unittest.mock import MagicMock, patch
from csle_common.controllers.flags_controller import FlagsController
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.dao.emulation_config.node_flags_config import NodeFlagsConfig
from csle_common.dao.emulation_config.flag import Flag


class TestFlagsControllerSuite:
    """
    Test suite for flags controller
    """

    @patch("csle_common.util.emulation_util.EmulationUtil.connect_admin")
    @patch("csle_common.util.emulation_util.EmulationUtil.execute_ssh_cmd")
    @patch("csle_common.util.emulation_util.EmulationUtil.disconnect_admin")
    def test_create_flags(self, mock_disconnect_admin, mock_execute_ssh_cmd, mock_connect_admin) -> None:
        """
        Test case for creating flags using the FlagsController

        :param mock_disconnect_admin: Mocked disconnect_admin method
        :param mock_execute_ssh_cmd: Mocked execute_ssh_cmd method
        :param mock_connect_admin: Mocked connect_admin
        :return: None
        """
        emulation_env_config = MagicMock(spec=EmulationEnvConfig)
        flags_config = MagicMock()
        emulation_env_config.flags_config = flags_config
        flags_config.node_flag_configs = [
            NodeFlagsConfig(
                ip="10.0.0.1",
                docker_gw_bridge_ip="172.17.0.1",
                physical_host_ip="192.168.0.1",
                flags=[
                    Flag(name="flag1", dir="/dir1", id=1, path="/path/to/flag1"),
                    Flag(name="flag2", dir="/dir2", id=2, path="/path/to/flag2"),
                ]
            )
        ]
        logger = MagicMock(spec=logging.Logger)
        physical_server_ip = "192.168.0.1"
        FlagsController.create_flags(emulation_env_config, physical_server_ip, logger)
        mock_connect_admin.assert_called_once_with(emulation_env_config=emulation_env_config, ip="172.17.0.1")
        mock_disconnect_admin.assert_called_once_with(emulation_env_config=emulation_env_config)
        mock_execute_ssh_cmd.assert_called()



from typing import List, Dict, Any, Union
import logging
import time
import subprocess
import random
import paramiko
import csle_collector.constants.constants as collector_constants
import csle_ryu.constants.constants as ryu_constants
import csle_common.constants.constants as constants
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.dao.emulation_config.node_resources_config import NodeResourcesConfig
from csle_common.controllers.container_controller import ContainerController
from csle_common.controllers.snort_ids_controller import SnortIDSController
from csle_common.controllers.ossec_ids_controller import OSSECIDSController
from csle_common.controllers.host_controller import HostController
from csle_common.controllers.kafka_controller import KafkaController
from csle_common.controllers.elk_controller import ELKController
from csle_common.controllers.sdn_controller_manager import SDNControllerManager
from csle_common.controllers.traffic_controller import TrafficController
from csle_common.controllers.five_g_core_controller import FiveGCoreController
from csle_common.controllers.five_g_cu_controller import FiveGCUController
from csle_common.controllers.five_g_du_controller import FiveGDUController
from csle_common.util.emulation_util import EmulationUtil
from csle_common.metastore.metastore_facade import MetastoreFacade
from csle_common.util.experiment_util import ExperimentUtil
from csle_common.logging.log import Logger
from csle_common.dao.emulation_config.emulation_execution import EmulationExecution
from csle_common.dao.emulation_config.emulation_execution_info import EmulationExecutionInfo
from csle_common.dao.emulation_config.config import Config
from csle_common.tunneling.forward_tunnel_thread import ForwardTunnelThread
from csle_common.util.cluster_util import ClusterUtil


class EmulationEnvController:
    """
    Class managing emulation environments
    """

    @staticmethod
    def stop_all_executions_of_emulation(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                                         logger: logging.Logger) -> None:
        """
        Stops all executions of a given emulation

        :param emulation_env_config: the emulation for which executions should be stopped
        :param physical_server_ip: ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        executions = MetastoreFacade.list_emulation_executions_for_a_given_emulation(
            emulation_name=emulation_env_config.name)
        for exec in executions:
            EmulationEnvController.stop_containers(execution=exec, physical_server_ip=physical_server_ip,
                                                   logger=logger)
            ContainerController.stop_docker_stats_thread(execution=exec, physical_server_ip=physical_server_ip,
                                                         logger=logger)

    @staticmethod
    def stop_execution_of_emulation(emulation_env_config: EmulationEnvConfig, execution_id: int,
                                    physical_server_ip: str, logger: logging.Logger) -> None:
        """
        Stops an execution of a given emulation

        :param emulation_env_config: the emulation for which executions should be stopped
        :param execution_id: the id of the execution to stop
        :param physical_server_ip: ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        execution = MetastoreFacade.get_emulation_execution(emulation_name=emulation_env_config.name,
                                                            ip_first_octet=execution_id)
        if execution is None:
            raise ValueError(f"Could not find any execution with id: {execution_id}, "
                             f"emulation: {emulation_env_config.name}")
        EmulationEnvController.stop_containers(execution=execution, physical_server_ip=physical_server_ip,
                                               logger=logger)
        ContainerController.stop_docker_stats_thread(execution=execution, physical_server_ip=physical_server_ip,
                                                     logger=logger)

    @staticmethod
    def stop_all_executions(physical_server_ip: str, logger: logging.Logger) -> None:
        """
        Stops all emulation executions

        :param physical_server_ip: ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        executions = MetastoreFacade.list_emulation_executions()
        for exec in executions:
            EmulationEnvController.stop_containers(execution=exec, physical_server_ip=physical_server_ip,
                                                   logger=logger)
            ContainerController.stop_docker_stats_thread(execution=exec, physical_server_ip=physical_server_ip,
                                                         logger=logger)

    @staticmethod
    def install_csle_collector_and_ryu_libraries(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                                                 logger: logging.Logger) \
            -> None:
        """
        Installs the latest csle-collector and csle-ryu libraries on all nodes of a given emulation

        :param emulation_env_config: the emulation configuration
        :param physical_server_ip: the IP of the physical servers where the containers are
        :param logger: the logger to use for logging
        :return: None
        """
        containers = list(filter(lambda x: x.physical_host_ip == physical_server_ip,
                                 emulation_env_config.containers_config.containers))
        ips = list(map(lambda x: x.docker_gw_bridge_ip, containers))
        if emulation_env_config.kafka_config.container.physical_host_ip == physical_server_ip:
            ips.append(emulation_env_config.kafka_config.container.docker_gw_bridge_ip)
        if emulation_env_config.elk_config.container.physical_host_ip == physical_server_ip:
            ips.append(emulation_env_config.elk_config.container.docker_gw_bridge_ip)
        if emulation_env_config.sdn_controller_config is not None:
            ips.append(emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip)
        for ip in ips:
            logger.info(f"Installing csle-collector version "
                        f"{emulation_env_config.csle_collector_version} on node: {ip}")
            EmulationUtil.connect_admin(emulation_env_config=emulation_env_config, ip=ip)
            cmd = collector_constants.INSTALL
            if emulation_env_config.csle_collector_version != collector_constants.LATEST_VERSION:
                cmd = cmd + f"=={emulation_env_config.csle_collector_version}"
            o, e, _ = EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip))
            time.sleep(2)
            o, e, _ = EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip))
            time.sleep(2)
            logger.info(f"Installing csle-ryu version "
                        f"{emulation_env_config.csle_ryu_version} on node: {ip}")
            EmulationUtil.connect_admin(emulation_env_config=emulation_env_config, ip=ip)
            cmd = ryu_constants.INSTALL
            if emulation_env_config.csle_ryu_version != ryu_constants.LATEST_VERSION:
                cmd = cmd + f"=={emulation_env_config.csle_ryu_version}"
            o, e, _ = EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip))
            time.sleep(2)
            o, e, _ = EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip))

            EmulationUtil.disconnect_admin(emulation_env_config=emulation_env_config)

    @staticmethod
    def update_execution_config_w_docker_gw_bridge_ip(execution: EmulationExecution) -> EmulationExecution:
        """
        Updates the execution configuration with the IP of the docker gw of the docker swarm

        :param execution: the execution to update
        :return: the updated execution
        """
        emulation_env_config = execution.emulation_env_config
        emulation_env_config.kafka_config.resources.docker_gw_bridge_ip = \
            emulation_env_config.kafka_config.container.docker_gw_bridge_ip
        emulation_env_config.kafka_config.resources.physical_host_ip = \
            emulation_env_config.kafka_config.container.physical_host_ip
        emulation_env_config.elk_config.resources.docker_gw_bridge_ip = \
            emulation_env_config.elk_config.container.docker_gw_bridge_ip
        emulation_env_config.elk_config.resources.physical_host_ip = \
            emulation_env_config.elk_config.container.physical_host_ip
        if emulation_env_config.sdn_controller_config is not None:
            emulation_env_config.sdn_controller_config.resources.docker_gw_bridge_ip = \
                emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip
            emulation_env_config.sdn_controller_config.resources.physical_host_ip = \
                emulation_env_config.sdn_controller_config.container.physical_host_ip
        for resource_config in emulation_env_config.resources_config.node_resources_configurations:
            for container in emulation_env_config.containers_config.containers:
                if container.get_readable_name() == resource_config.container_name:
                    resource_config.docker_gw_bridge_ip = container.docker_gw_bridge_ip
                    resource_config.physical_host_ip = container.physical_host_ip
        for ovs_switch in emulation_env_config.ovs_config.switch_configs:
            for container in emulation_env_config.containers_config.containers:
                if container.get_readable_name() == ovs_switch.container_name:
                    ovs_switch.docker_gw_bridge_ip = container.docker_gw_bridge_ip
                    ovs_switch.physical_host_ip = container.physical_host_ip
        for user_config in emulation_env_config.users_config.users_configs:
            for container in emulation_env_config.containers_config.containers:
                if user_config.ip in container.get_ips():
                    user_config.docker_gw_bridge_ip = container.docker_gw_bridge_ip
                    user_config.physical_host_ip = container.physical_host_ip
        for vuln_config in emulation_env_config.vuln_config.node_vulnerability_configs:
            for container in emulation_env_config.containers_config.containers:
                if vuln_config.ip in container.get_ips():
                    vuln_config.docker_gw_bridge_ip = container.docker_gw_bridge_ip
                    vuln_config.physical_host_ip = container.physical_host_ip
        for flags_config in emulation_env_config.flags_config.node_flag_configs:
            for container in emulation_env_config.containers_config.containers:
                if flags_config.ip in container.get_ips():
                    flags_config.docker_gw_bridge_ip = container.docker_gw_bridge_ip
                    flags_config.physical_host_ip = container.physical_host_ip
        for node_fw_config in emulation_env_config.topology_config.node_configs:
            for container in emulation_env_config.containers_config.containers:
                for ip in container.get_ips():
                    if ip in node_fw_config.get_ips():
                        node_fw_config.docker_gw_bridge_ip = container.docker_gw_bridge_ip
                        node_fw_config.physical_host_ip = container.physical_host_ip
                        break
        emulation_env_config.kafka_config.firewall_config.docker_gw_bridge_ip = \
            emulation_env_config.kafka_config.container.docker_gw_bridge_ip
        emulation_env_config.kafka_config.firewall_config.physical_host_ip = \
            emulation_env_config.kafka_config.container.physical_host_ip
        emulation_env_config.elk_config.firewall_config.docker_gw_bridge_ip = \
            emulation_env_config.elk_config.container.docker_gw_bridge_ip
        emulation_env_config.elk_config.firewall_config.physical_host_ip = \
            emulation_env_config.elk_config.container.physical_host_ip
        if emulation_env_config.sdn_controller_config is not None:
            emulation_env_config.sdn_controller_config.firewall_config.docker_gw_bridge_ip = \
                emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip
            emulation_env_config.sdn_controller_config.firewall_config.physical_host_ip = \
                emulation_env_config.sdn_controller_config.container.physical_host_ip
        for node_traffic_config in emulation_env_config.traffic_config.node_traffic_configs:
            for container in emulation_env_config.containers_config.containers:
                if node_traffic_config.ip in container.get_ips():
                    node_traffic_config.docker_gw_bridge_ip = container.docker_gw_bridge_ip
                    node_traffic_config.physical_host_ip = container.physical_host_ip
        for container in emulation_env_config.containers_config.containers:
            if emulation_env_config.traffic_config.client_population_config.ip in container.get_ips():
                emulation_env_config.traffic_config.client_population_config.docker_gw_bridge_ip = \
                    container.docker_gw_bridge_ip
                emulation_env_config.traffic_config.client_population_config.physical_host_ip = \
                    container.physical_host_ip
        execution.emulation_env_config = emulation_env_config
        MetastoreFacade.update_emulation_execution(emulation_execution=execution,
                                                   ip_first_octet=execution.ip_first_octet,
                                                   emulation=execution.emulation_name)
        return execution

    @staticmethod
    def apply_kafka_config(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                           logger: logging.Logger) -> None:
        """
        Applies the kafka config

        :param emulation_env_config: the emulation env config
        :param physical_server_ip: ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        if emulation_env_config.kafka_config.container.physical_host_ip != physical_server_ip:
            return
        steps = 3
        current_step = 1
        logger.info("-- Configuring the kafka container --")

        logger.info(
            f"-- Kafka configuration step {current_step}/{steps}: Configuring the IP addresses of the kafka brokers --")
        KafkaController.configure_broker_ips(emulation_env_config=emulation_env_config, logger=logger)
        current_step += 1

        logger.info(
            f"-- Kafka configuration step {current_step}/{steps}: Restarting the Kafka server --")
        KafkaController.stop_kafka_server(emulation_env_config=emulation_env_config, logger=logger)
        time.sleep(20)
        KafkaController.start_kafka_server(emulation_env_config=emulation_env_config, logger=logger)
        time.sleep(20)

        current_step += 1
        logger.info(f"-- Kafka configuration step {current_step}/{steps}: Create topics --")
        KafkaController.create_topics(emulation_env_config=emulation_env_config, logger=logger)

    @staticmethod
    def start_custom_traffic(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                             no_traffic: bool = True) -> None:
        """
        Utility function for starting traffic generators and client population on a given emulation

        :param emulation_env_config: the configuration of the emulation
        :param no_traffic boolean flag whether the internal traffic generators should be skipped.
        :param physical_server_ip: ip of the physical servern
        :return: None
        """
        if not no_traffic:
            TrafficController.start_internal_traffic_generators(emulation_env_config=emulation_env_config,
                                                                physical_server_ip=physical_server_ip,
                                                                logger=Logger.__call__().get_logger())
        TrafficController.stop_client_producer(emulation_env_config=emulation_env_config,
                                               physical_server_ip=physical_server_ip,
                                               logger=Logger.__call__().get_logger())
        TrafficController.start_client_population(emulation_env_config=emulation_env_config,
                                                  physical_server_ip=physical_server_ip,
                                                  logger=Logger.__call__().get_logger())
        TrafficController.start_client_producer(emulation_env_config=emulation_env_config,
                                                physical_server_ip=physical_server_ip,
                                                logger=Logger.__call__().get_logger())

    @staticmethod
    def delete_networks_of_emulation_env_config(emulation_env_config: EmulationEnvConfig,
                                                physical_server_ip: str, logger: logging.Logger,
                                                leader: bool = False) -> None:
        """
        Deletes the docker networks

        :param emulation_env_config: the emulation env config
        :param physical_server_ip: the ip of the physical server to remove the networks
        :param leader: boolean flag indicating whether this node is the leader in the Swarm cluster
        :param logger: the logger to use for logging
        :return: None
        """
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip == physical_server_ip or leader:
                for ip_net in c.ips_and_networks:
                    ip, net = ip_net
                    ContainerController.remove_network(name=net.name, logger=logger)

        c = emulation_env_config.kafka_config.container
        if c.physical_host_ip == physical_server_ip or leader:
            for ip_net in c.ips_and_networks:
                ip, net = ip_net
                ContainerController.remove_network(name=net.name, logger=logger)

    @staticmethod
    def create_execution(emulation_env_config: EmulationEnvConfig, physical_servers: List[str], logger: logging.Logger,
                         id: int = -1) -> EmulationExecution:
        """
        Creates a new emulation execution

        :param emulation_env_config: the emulation configuration
        :param physical_servers: the physical servers to deploy the containers on
        :param id: the id of the execution (if not specified the next available id will be used)
        :param logger: the logger to use for logging
        :return: a DTO representing the execution
        """
        timestamp = float(time.time())
        total_subnets = constants.CSLE.LIST_OF_IP_SUBNETS
        used_subnets = list(map(lambda x: x.ip_first_octet,
                                MetastoreFacade.list_emulation_executions_for_a_given_emulation(
                                    emulation_name=emulation_env_config.name)))
        available_subnets = list(filter(lambda x: x not in used_subnets, total_subnets))
        ip_first_octet = available_subnets[0]
        if id != -1 and id not in available_subnets:
            logger.warning(f"The specified execution ID: {id} is not valid or is already taken. "
                           f"Using ID: {ip_first_octet} instead")
        elif id != -1 and id in available_subnets:
            ip_first_octet = id

        em_config = emulation_env_config.create_execution_config(ip_first_octet=ip_first_octet,
                                                                 physical_servers=physical_servers)
        emulation_execution = EmulationExecution(emulation_name=emulation_env_config.name,
                                                 timestamp=timestamp, ip_first_octet=ip_first_octet,
                                                 emulation_env_config=em_config, physical_servers=physical_servers)
        MetastoreFacade.save_emulation_execution(emulation_execution=emulation_execution)
        return emulation_execution

    @staticmethod
    def run_containers(emulation_execution: EmulationExecution, physical_host_ip: str, logger: logging.Logger) -> None:
        """
        Run containers in the emulation env config

        :param emulation_execution: the execution DTO
        :param physical_host_ip: the ip of the physical host where the containers should be started
        :param logger: the logger to use for logging
        :return: None
        """
        path = ExperimentUtil.default_output_dir()
        emulation_env_config = emulation_execution.emulation_env_config

        # Start regular containers
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_host_ip:
                continue
            container_resources: Union[None, NodeResourcesConfig] = None
            for r in emulation_env_config.resources_config.node_resources_configurations:
                if r.container_name == c.get_readable_name():
                    container_resources = r
            if container_resources is None:
                raise ValueError(f"Container resources not found for container with name:{c.get_readable_name()}, "
                                 f"resources:{emulation_env_config.resources_config}")
            name = c.get_full_name()
            cmd = f"docker container run -dt --name {name} " \
                  f"--hostname={c.name}{c.suffix} --label dir={path} " \
                  f"--label cfg={path + constants.DOCKER.EMULATION_ENV_CFG_PATH} " \
                  f"-e TZ=Europe/Stockholm " \
                  f"--label emulation={emulation_env_config.name} --network=none --publish-all=true " \
                  f"--memory={container_resources.available_memory_gb}G --cpus={container_resources.num_cpus} " \
                  f"--restart={c.restart_policy} --cap-add NET_ADMIN --cap-add=SYS_NICE --privileged " \
                  f"{constants.CONTAINER_IMAGES.DOCKERHUB_USERNAME}/{c.name}:{c.version}"
            logger.info(f"Starting container:{name} with cmd: {cmd}")
            subprocess.call(cmd, shell=True)

        if emulation_env_config.kafka_config.container.physical_host_ip == physical_host_ip:
            # Start the kafka container
            c = emulation_env_config.kafka_config.container
            container_resources = emulation_env_config.kafka_config.resources
            name = c.get_full_name()
            cmd = f"docker container run -dt --name {name} " \
                  f"--hostname={c.name}{c.suffix} --label dir={path} " \
                  f"--label cfg={path + constants.DOCKER.EMULATION_ENV_CFG_PATH} " \
                  f"-e TZ=Europe/Stockholm " \
                  f"--label emulation={emulation_env_config.name} --network=none --publish-all=true " \
                  f"--memory={container_resources.available_memory_gb}G --cpus={container_resources.num_cpus} " \
                  f"--restart={c.restart_policy} --cap-add NET_ADMIN --cap-add=SYS_NICE --privileged " \
                  f"{constants.CONTAINER_IMAGES.DOCKERHUB_USERNAME}/{c.name}:{c.version}"
            logger.info(f"Starting container:{name}, cmd: {cmd}")
            subprocess.call(cmd, shell=True)

        if emulation_env_config.elk_config.container.physical_host_ip == physical_host_ip:
            # Start the ELK container
            c = emulation_env_config.elk_config.container
            container_resources = emulation_env_config.elk_config.resources
            name = c.get_full_name()
            cmd = f"docker container run -dt --name {name} " \
                  f"--hostname={c.name}{c.suffix} --label dir={path} " \
                  f"--label cfg={path + constants.DOCKER.EMULATION_ENV_CFG_PATH} " \
                  f"-e TZ=Europe/Stockholm " \
                  f"--label emulation={emulation_env_config.name} --network=none --publish-all=true " \
                  f"--memory={container_resources.available_memory_gb}G --cpus={container_resources.num_cpus} " \
                  f"--restart={c.restart_policy} --cap-add NET_ADMIN --cap-add=SYS_NICE --privileged " \
                  f"{constants.CONTAINER_IMAGES.DOCKERHUB_USERNAME}/{c.name}:{c.version}"
            logger.info(f"Starting container:{name}, cmd: {cmd}")
            subprocess.call(cmd, shell=True)

        if emulation_env_config.sdn_controller_config is not None \
                and emulation_env_config.sdn_controller_config.container.physical_host_ip == physical_host_ip:
            # Start the SDN controller container
            c = emulation_env_config.sdn_controller_config.container
            container_resources = emulation_env_config.sdn_controller_config.resources
            name = c.get_full_name()
            cmd = f"docker container run -dt --name {name} " \
                  f"--hostname={c.name}{c.suffix} --label dir={path} " \
                  f"--label cfg={path + constants.DOCKER.EMULATION_ENV_CFG_PATH} " \
                  f"-e TZ=Europe/Stockholm " \
                  f"--label emulation={emulation_env_config.name} --network=none --publish-all=true " \
                  f"--memory={container_resources.available_memory_gb}G --cpus={container_resources.num_cpus} " \
                  f"--restart={c.restart_policy} --cap-add NET_ADMIN --cap-add=SYS_NICE --privileged " \
                  f"{constants.CONTAINER_IMAGES.DOCKERHUB_USERNAME}/{c.name}:{c.version}"
            logger.info(f"Starting container:{name}, cmd: {cmd}")
            subprocess.call(cmd, shell=True)

    @staticmethod
    def start_containers_of_execution(emulation_execution: EmulationExecution, physical_host_ip: str) -> None:
        """
        Starts stopped containers in a given emulation execution

        :param emulation_execution: the execution DTO
        :param physical_host_ip: the ip of the physical host
        :return: None
        """
        emulation_env_config = emulation_execution.emulation_env_config

        # Start regular containers
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip == physical_host_ip:
                ContainerController.start_container(name=c.get_full_name())

        # Start the kafka container
        c = emulation_env_config.kafka_config.container
        if c.physical_host_ip == physical_host_ip:
            ContainerController.start_container(name=c.get_full_name())

        # Start the ELK container
        c = emulation_env_config.elk_config.container
        if c.physical_host_ip == physical_host_ip:
            ContainerController.start_container(name=c.get_full_name())

        if emulation_env_config.sdn_controller_config is not None:
            # Start the SDN controller container
            c = emulation_env_config.sdn_controller_config.container
            if c.physical_host_ip == physical_host_ip:
                ContainerController.start_container(name=c.get_full_name())

    @staticmethod
    def run_container(image: str, name: str, logger: logging.Logger, memory: int = 4, num_cpus: int = 1,
                      create_network: bool = True, version: str = "0.0.1") -> None:
        """
        Runs a given container

        :param image: image of the container
        :param name: name of the container
        :param memory: memory in GB
        :param num_cpus: number of CPUs to allocate
        :param create_network: whether to create a virtual network or not
        :param version: the version tag
        :param logger: the logger to use for logging
        :return: None
        """
        logger.info(f"Starting container with image:{image} and name:csle_{name}-{version.replace('.', '')}")
        if create_network:
            net_id = random.randint(128, 254)
            sub_net_id = random.randint(2, 254)
            host_id = random.randint(2, 254)
            net_name = f"csle_custom_net_{name}_{net_id}"
            ip = f"10.{net_id}.{sub_net_id}.{host_id}"
            ContainerController.create_network(
                name=net_name, subnetmask=f"10.{net_id}.0.0/16", existing_network_names=[], logger=logger)
            cmd = f"docker container run -dt --name csle_{name}-{version.replace('.', '')} " \
                  f"--hostname={name} " \
                  f"-e TZ=Europe/Stockholm " \
                  f"--network={net_name} --ip {ip} --publish-all=true " \
                  f"--memory={memory}G --cpus={num_cpus} " \
                  f"--restart={constants.DOCKER.ON_FAILURE_3} --cap-add NET_ADMIN --privileged " \
                  f"--cap-add=SYS_NICE {image}"
        else:
            cmd = f"docker container run -dt --name csle-{name}-{version.replace('.', '')} " \
                  f"--hostname={name} " \
                  f"-e TZ=Europe/Stockholm --net=none " \
                  f"--publish-all=true " \
                  f"--memory={memory}G --cpus={num_cpus} " \
                  f"--restart={constants.DOCKER.ON_FAILURE_3} --cap-add NET_ADMIN --privileged " \
                  f"--cap-add=SYS_NICE {image}"
        subprocess.call(cmd, shell=True)

    @staticmethod
    def stop_containers(execution: EmulationExecution, physical_server_ip: str, logger: logging.Logger) -> None:
        """
        Stop containers in the emulation env config

        :param execution: the execution to stop
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        emulation_env_config = execution.emulation_env_config

        # Stop regular containers
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            name = c.get_full_name()
            logger.info(f"Stopping container:{name}")
            cmd = f"docker stop {name}"
            subprocess.call(cmd, shell=True)

        # Stop the Kafka container
        c = emulation_env_config.kafka_config.container
        if c.physical_host_ip == physical_server_ip:
            name = c.get_full_name()
            logger.info(f"Stopping container:{name}")
            cmd = f"docker stop {name}"
            subprocess.call(cmd, shell=True)

        # Stop the ELK container
        c = emulation_env_config.elk_config.container
        if c.physical_host_ip == physical_server_ip:
            name = c.get_full_name()
            logger.info(f"Stopping container:{name}")
            cmd = f"docker stop {name}"
            subprocess.call(cmd, shell=True)

        if emulation_env_config.sdn_controller_config is not None:
            # Stop the SDN controller container
            c = emulation_env_config.sdn_controller_config.container
            if c.physical_host_ip == physical_server_ip:
                name = c.get_full_name()
                logger.info(f"Stopping container:{name}")
                cmd = f"docker stop {name}"
                subprocess.call(cmd, shell=True)

    @staticmethod
    def clean_all_emulation_executions(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                                       logger: logging.Logger, leader: bool = False) -> None:
        """
        Cleans an emulation

        :param emulation_env_config: the config of the emulation to clean
        :param physical_server_ip: the ip of the physical server to clean the emulation executions
        :param leader: boolean flag indicating whether this node is the leader in the Swarm cluster or not
        :param logger: the logger to use for logging
        :return: None
        """
        executions = MetastoreFacade.list_emulation_executions_for_a_given_emulation(
            emulation_name=emulation_env_config.name)
        for exec in executions:
            EmulationEnvController.stop_containers(execution=exec, physical_server_ip=physical_server_ip,
                                                   logger=logger)
            EmulationEnvController.rm_containers(execution=exec, physical_server_ip=physical_server_ip, logger=logger)
            try:
                ContainerController.stop_docker_stats_thread(execution=exec, physical_server_ip=physical_server_ip,
                                                             logger=logger)
            except Exception:
                pass
            EmulationEnvController.delete_networks_of_emulation_env_config(
                emulation_env_config=exec.emulation_env_config, physical_server_ip=physical_server_ip, logger=logger,
                leader=leader)

    @staticmethod
    def clean_emulation_execution(emulation_env_config: EmulationEnvConfig, execution_id: int,
                                  physical_server_ip: str, logger: logging.Logger, leader: bool = False) -> None:
        """
        Cleans an emulation execution

        :param execution_id: the id of the execution to clean
        :param emulation_env_config: the config of the emulation to clean
        :param physical_server_ip: the ip of the physical server to clean the execution
        :param leader: boolean flag indicating whether this node is the leader or not
        :param logger: the logger to use for logging
        :return: None
        """
        execution = MetastoreFacade.get_emulation_execution(ip_first_octet=execution_id,
                                                            emulation_name=emulation_env_config.name)
        if execution is None:
            raise ValueError(f"Could not find any execution with id: {execution_id}, "
                             f"emulation: {emulation_env_config.name}")
        EmulationEnvController.stop_containers(execution=execution, physical_server_ip=physical_server_ip,
                                               logger=logger)
        EmulationEnvController.rm_containers(execution=execution, physical_server_ip=physical_server_ip, logger=logger)
        try:
            ContainerController.stop_docker_stats_thread(execution=execution, physical_server_ip=physical_server_ip,
                                                         logger=logger)
        except Exception:
            pass
        EmulationEnvController.delete_networks_of_emulation_env_config(
            emulation_env_config=execution.emulation_env_config, physical_server_ip=physical_server_ip, logger=logger,
            leader=leader)

    @staticmethod
    def clean_all_executions(physical_server_ip: str, logger: logging.Logger, leader: bool = False) -> None:
        """
        Cleans all executions of a given emulation on a given physical server

        :param physical_server_ip: the ip of the physical server to clean the executions
        :param logger: the logger to use for logging
        :param leader: boolean flag indicating whether this node is the leader or not
        :return: None
        """
        executions = MetastoreFacade.list_emulation_executions()
        for exec in executions:
            EmulationEnvController.stop_containers(execution=exec, physical_server_ip=physical_server_ip,
                                                   logger=logger)
            EmulationEnvController.rm_containers(execution=exec, physical_server_ip=physical_server_ip, logger=logger)
            try:
                ContainerController.stop_docker_stats_thread(execution=exec, physical_server_ip=physical_server_ip,
                                                             logger=logger)
            except Exception:
                pass
            EmulationEnvController.delete_networks_of_emulation_env_config(
                emulation_env_config=exec.emulation_env_config, physical_server_ip=physical_server_ip, logger=logger,
                leader=leader)
            MetastoreFacade.remove_emulation_execution(emulation_execution=exec)

    @staticmethod
    def rm_containers(execution: EmulationExecution, physical_server_ip: str, logger: logging.Logger) -> None:
        """
        Remove containers in the emulation env config for a given execution

        :param execution: the execution to remove
        :param physical_server_ip: the ip of the physical server to remove the containers
        :param logger: the logger to use for logging
        :return: None
        """

        # Remove regular containers
        for c in execution.emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            name = c.get_full_name()
            logger.info(f"Removing container:{name}")
            cmd = f"docker rm {name}"
            subprocess.call(cmd, shell=True)

        # Remove the kafka container
        c = execution.emulation_env_config.kafka_config.container
        if c.physical_host_ip == physical_server_ip:
            name = c.get_full_name()
            logger.info(f"Removing container:{name}")
            cmd = f"docker rm {name}"
            subprocess.call(cmd, shell=True)

        # Remove the elk container
        c = execution.emulation_env_config.elk_config.container
        if c.physical_host_ip == physical_server_ip:
            name = c.get_full_name()
            logger.info(f"Removing container:{name}")
            cmd = f"docker rm {name}"
            subprocess.call(cmd, shell=True)

        if execution.emulation_env_config.sdn_controller_config is not None:
            # Remove the SDN controller container
            c = execution.emulation_env_config.sdn_controller_config.container
            if c.physical_host_ip == physical_server_ip:
                name = c.get_full_name()
                logger.info(f"Removing container:{name}")
                cmd = f"docker rm {name}"
                subprocess.call(cmd, shell=True)

    @staticmethod
    def install_emulation(config: EmulationEnvConfig) -> None:
        """
        Installs the emulation configuration in the metastore

        :param config: the config to install
        :return: None
        """
        MetastoreFacade.install_emulation(config=config)

    @staticmethod
    def save_emulation_image(img: bytes, emulation_name: str) -> None:
        """
        Saves the emulation image

        :param image: the image data
        :param emulation_name: the name of the emulation
        :return: None
        """
        MetastoreFacade.save_emulation_image(img=img, emulation_name=emulation_name)

    @staticmethod
    def uninstall_emulation(config: EmulationEnvConfig) -> None:
        """
        Uninstalls the emulation configuration in the metastore

        :param config: the config to uninstall
        :return: None
        """
        MetastoreFacade.uninstall_emulation(config=config)

    @staticmethod
    def ping_all(emulation_env_config: EmulationEnvConfig, physical_server_ip: str, logger: logging.Logger) -> None:
        """
        Tests the connections between all the containers using ping

        :param emulation_env_config: the emulation config
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        if emulation_env_config.sdn_controller_config is not None \
                and emulation_env_config.sdn_controller_config.container.physical_host_ip == physical_server_ip:

            # Ping controller-switches
            for ovs_sw in emulation_env_config.ovs_config.switch_configs:
                logger.info(f"Ping {ovs_sw.controller_ip} to {ovs_sw.ip}")
                cmd = f"{constants.COMMANDS.DOCKER_EXEC_COMMAND} " \
                      f"{emulation_env_config.sdn_controller_config.container.get_full_name()} " \
                      f"{constants.COMMANDS.PING} " \
                      f"{ovs_sw.ip} -c 5 &"
                subprocess.Popen(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, shell=True)

                logger.info(f"Ping {ovs_sw.ip} to {ovs_sw.controller_ip}")
                cmd = f"{constants.COMMANDS.DOCKER_EXEC_COMMAND} {ovs_sw.container_name} {constants.COMMANDS.PING} " \
                      f"{ovs_sw.controller_ip} -c 5 &"
                subprocess.Popen(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, shell=True)

        # Ping containers to switches
        for c1 in emulation_env_config.containers_config.containers:
            if c1.physical_host_ip != physical_server_ip:
                continue
            for c2 in emulation_env_config.containers_config.containers:
                for ip in c2.get_ips():
                    logger.info(f"Ping {c1.get_ips()[0]} to {ip}")
                    cmd = f"{constants.COMMANDS.DOCKER_EXEC_COMMAND} {c1.get_full_name()} {constants.COMMANDS.PING} " \
                          f"{ip} -c 5 &"
                    subprocess.Popen(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, shell=True)

    @staticmethod
    def get_execution_info(execution: EmulationExecution, logger: logging.Logger, physical_server_ip: str) \
            -> EmulationExecutionInfo:
        """
        Gets runtime information about an execution

        :param execution: the emulation execution to get the information for
        :param logger: the logger to use for logging
        :param physical_server_ip: the IP of the physical server
        :return: execution information
        """
        running_containers, stopped_containers = ContainerController.list_all_running_containers_in_emulation(
            emulation_env_config=execution.emulation_env_config)
        active_ips: List[str] = []
        for container in running_containers:
            active_ips = active_ips + container.get_ips()
            active_ips.append(container.docker_gw_bridge_ip)
        active_ips.append(constants.COMMON.LOCALHOST)
        active_ips.append(constants.COMMON.LOCALHOST_127_0_0_1)
        active_ips.append(constants.COMMON.LOCALHOST_127_0_1_1)
        config = Config.get_current_config()
        if config is None:
            raise ValueError("Could not cluster read configuration")
        for node in config.cluster_config.cluster_nodes:
            active_ips.append(node.ip)
        emulation_name = execution.emulation_name
        execution_id = execution.ip_first_octet
        logger.info(f"Getting the Snort IDS Managers info for emulation: {emulation_name}, "
                    f"execution id: {execution_id}")
        snort_ids_managers_info = \
            SnortIDSController.get_snort_managers_info(emulation_env_config=execution.emulation_env_config,
                                                       active_ips=active_ips, logger=logger,
                                                       physical_server_ip=physical_server_ip)
        logger.info(f"Getting the OSSEC IDS Managers info for emulation: {emulation_name}, "
                    f"execution id: {execution_id}")
        ossec_ids_managers_info = \
            OSSECIDSController.get_ossec_managers_info(emulation_env_config=execution.emulation_env_config,
                                                       active_ips=active_ips, logger=logger,
                                                       physical_host_ip=physical_server_ip)
        logger.info(f"Getting the Kafka Managers info for emulation: {emulation_name}, "
                    f"execution id: {execution_id}")
        kafka_managers_info = \
            KafkaController.get_kafka_managers_info(emulation_env_config=execution.emulation_env_config,
                                                    active_ips=active_ips, logger=logger,
                                                    physical_host_ip=physical_server_ip)
        logger.info(f"Getting the Host Managers info for emulation: {emulation_name}, "
                    f"execution id: {execution_id}")
        host_managers_info = \
            HostController.get_host_managers_info(emulation_env_config=execution.emulation_env_config,
                                                  active_ips=active_ips, logger=logger,
                                                  physical_host_ip=physical_server_ip)
        logger.info(f"Getting the Client Managers info for emulation: {emulation_name}, "
                    f"execution id: {execution_id}")
        client_managers_info = \
            TrafficController.get_client_managers_info(emulation_env_config=execution.emulation_env_config,
                                                       active_ips=active_ips, logger=logger)
        logger.info(f"Getting the Traffic Managers info for emulation: {emulation_name}, "
                    f"execution id: {execution_id}")
        traffic_managers_info = \
            TrafficController.get_traffic_managers_info(emulation_env_config=execution.emulation_env_config,
                                                        active_ips=active_ips, logger=logger,
                                                        physical_host_ip=physical_server_ip)
        logger.info(f"Getting the Docker Stats Managers info for emulation: {emulation_name}, "
                    f"execution id: {execution_id}")
        docker_stats_managers_info = \
            ContainerController.get_docker_stats_managers_info(emulation_env_config=execution.emulation_env_config,
                                                               active_ips=active_ips, logger=logger,
                                                               physical_host_ip=physical_server_ip)
        logger.info(f"Getting the Elk Managers info for emulation: {emulation_name}, "
                    f"execution id: {execution_id}")
        elk_managers_info = \
            ELKController.get_elk_managers_info(emulation_env_config=execution.emulation_env_config,
                                                active_ips=active_ips, logger=logger,
                                                physical_host_ip=physical_server_ip)
        logger.info("Getting the 5G Core Managers info")
        five_g_core_managers_info = \
            FiveGCoreController.get_five_g_core_managers_info(emulation_env_config=execution.emulation_env_config,
                                                              active_ips=active_ips, logger=logger,
                                                              physical_server_ip=physical_server_ip)
        logger.info("Getting the 5G CU Managers info")
        five_g_cu_managers_info = \
            FiveGCUController.get_five_g_cu_managers_info(emulation_env_config=execution.emulation_env_config,
                                                          active_ips=active_ips, logger=logger,
                                                          physical_server_ip=physical_server_ip)
        logger.info("Getting the 5G DU Managers info")
        five_g_du_managers_info = \
            FiveGDUController.get_five_g_du_managers_info(emulation_env_config=execution.emulation_env_config,
                                                          active_ips=active_ips, logger=logger,
                                                          physical_server_ip=physical_server_ip)
        active_networks, inactive_networks = ContainerController.list_all_active_networks_for_emulation(
            emulation_env_config=execution.emulation_env_config)
        ryu_managers_info = None
        if execution.emulation_env_config.sdn_controller_config is not None:
            ryu_managers_info = SDNControllerManager.get_ryu_managers_info(
                emulation_env_config=execution.emulation_env_config, active_ips=active_ips,
                logger=logger, physical_server_ip=physical_server_ip)
        execution_info = EmulationExecutionInfo(emulation_name=emulation_name, execution_id=execution_id,
                                                snort_ids_managers_info=snort_ids_managers_info,
                                                ossec_ids_managers_info=ossec_ids_managers_info,
                                                kafka_managers_info=kafka_managers_info,
                                                host_managers_info=host_managers_info,
                                                client_managers_info=client_managers_info,
                                                docker_stats_managers_info=docker_stats_managers_info,
                                                running_containers=running_containers,
                                                stopped_containers=stopped_containers,
                                                active_networks=active_networks,
                                                inactive_networks=inactive_networks,
                                                elk_managers_info=elk_managers_info,
                                                traffic_managers_info=traffic_managers_info,
                                                ryu_managers_info=ryu_managers_info,
                                                five_g_core_managers_info=five_g_core_managers_info,
                                                five_g_cu_managers_info=five_g_cu_managers_info,
                                                five_g_du_managers_info=five_g_du_managers_info)
        return execution_info

    @staticmethod
    def create_ssh_tunnel(tunnels_dict: Dict[str, Any], local_port: int,
                          remote_port: int, remote_ip: str, emulation: str, execution_id: int) -> None:
        """
        Creates an SSH tunnel to forward the port of a container

        :param execution: the emulation execution
        :param tunnels_dict: a dict with existing tunnels
        :param local_port: the local port to forward
        :param remote_port: the remote port to forward
        :param remote_ip: the remote ip to forward
        :param emulation: the name of the emulation
        :param execution_id: the id of the execution
        :return: None
        """
        config = Config.get_current_config()
        if config is None:
            ClusterUtil.set_config_parameters_from_config_file()
        config = Config.get_current_config()
        if config is None:
            raise ValueError("Could not read the CSLE configuration")
        conn = paramiko.SSHClient()
        if conn is None:
            raise ValueError("Could not create paramiko SSH client")
        conn.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        conn.connect(remote_ip, username=config.ssh_admin_username, password=config.ssh_admin_password)
        agent_transport = conn.get_transport()
        if agent_transport is None:
            raise ValueError(f"Error opening SSH connection to {remote_ip}")
        agent_transport.set_keepalive(5)
        tunnel_thread = ForwardTunnelThread(
            local_port=local_port,
            remote_host=remote_ip,
            remote_port=remote_port, transport=agent_transport,
            tunnels_dict=tunnels_dict)
        tunnel_thread.start()
        tunnel_thread_dict: Dict[str, Union[ForwardTunnelThread, int, str]] = {}
        tunnel_thread_dict[constants.GENERAL.THREAD_PROPERTY] = tunnel_thread
        tunnel_thread_dict[constants.GENERAL.PORT_PROPERTY] = local_port
        tunnel_thread_dict[constants.GENERAL.EMULATION_PROPERTY] = emulation
        tunnel_thread_dict[constants.GENERAL.EXECUTION_ID_PROPERTY] = execution_id
        tunnels_dict[remote_ip] = tunnel_thread_dict



from typing import Optional, List, Dict, Any
from csle_common.dao.emulation_config.transport_protocol import TransportProtocol
from csle_common.dao.emulation_config.node_vulnerability_config import NodeVulnerabilityConfig
from csle_common.dao.emulation_config.credential import Credential
from csle_common.dao.emulation_config.network_service import NetworkService
from csle_common.dao.emulation_config.vulnerability_type import VulnType
from csle_base.json_serializable import JSONSerializable


class EmulationVulnerabilityObservationState(JSONSerializable):
    """
    A DTO representing an observed vulnerability in the emulation
    """

    def __init__(self, name: str, port: int,
                 protocol: TransportProtocol, cvss: float, credentials: List[Credential],
                 osvdbid: int = -1,
                 description: Optional[str] = None,
                 service: Optional[str] = ""):
        """
        Initializes the DTO

        :param name: the name of the vulnerability
        :param port: the port of the vulnerability
        :param protocol: the protocol of the vulnerability
        :param cvss: the CVSS of the vulnerability
        :param osvdbid: the OSVDBID of the vulnerability
        :param description: the description of the vulnerability
        :param credentials: the credentials of the vulnerability
        :param service: the service of the vulnerability
        """
        self.name = name
        self.port = port
        self.protocol = protocol
        self.cvss = cvss
        self.osvdb_id = osvdbid
        self.description = description
        self.credentials = credentials
        self.service = service
        if service is None or service == "":
            for cr in self.credentials:
                if cr.service is not None and cr.service != "":
                    self.service = cr.service

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation
        
        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["name"] = self.name
        d["protocol"] = self.protocol
        d["cvss"] = self.cvss
        d["osvdb_id"] = self.osvdb_id
        d["description"] = self.description
        d["credentials"] = list(map(lambda x: x.to_dict(), self.credentials))
        d["service"] = self.service
        d["port"] = self.port
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "EmulationVulnerabilityObservationState":
        """
        Converts a dict representation of the object into an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = EmulationVulnerabilityObservationState(
            name=d["name"],
            protocol=d["protocol"],
            cvss=d["cvss"],
            osvdbid=d["osvdb_id"],
            description=d["description"],
            credentials=list(map(lambda x: Credential.from_dict(x), d["credentials"])),
            service=d["service"],
            port=d["port"]
        )
        return obj

    def __str__(self):
        """
        :return: a string representation of the observed vulnerability
        """
        return "name:{}, port:{}, protocol:{}, cvss:{}, osvdb_id:{}, desc:{}, service:{}, credentials:{}".format(
            self.name, self.port, self.protocol, self.cvss, self.osvdb_id, self.description, self.service,
            list(map(lambda x: str(x), self.credentials)) if self.credentials is not None else self.credentials)

    def to_vulnerability(self) -> NodeVulnerabilityConfig:
        """
        Converts the object into a vulnerability representation

        :return: A vulnerability representation of the object
        """
        vuln = NodeVulnerabilityConfig(name=self.name, port=self.port, protocol=self.protocol, cvss=self.cvss,
                                       cve="", credentials=self.credentials, service=self.service, ip="",
                                       vuln_type=VulnType.WEAK_PW)
        return vuln

    def to_network_services(self) -> List[NetworkService]:
        """
        Converts the object into a network service representation

        :return: the network service representation
        """
        services = [NetworkService(protocol=self.protocol if self.protocol is not None else "",
                                   port=self.port, name=self.service if self.service is not None else "",
                                   credentials=self.credentials)]
        for cr in self.credentials:
            new_service = True
            for s in services:
                if s.name == cr.service:
                    new_service = False
            if new_service:
                services.append(NetworkService.from_credential(cr))
        return services

    @staticmethod
    def from_json_file(json_file_path: str) -> "EmulationVulnerabilityObservationState":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return EmulationVulnerabilityObservationState.from_dict(json.loads(json_str))

    def num_attributes(self) -> int:
        """
        :return: The number of attribute of the DTO
        """
        if len(self.credentials) > 0:
            return 7 + len(self.credentials) * self.credentials[0].num_attributes()
        else:
            return 0

    @staticmethod
    def schema() -> "EmulationVulnerabilityObservationState":
        """
        :return: get the schema of the DTO
        """
        return EmulationVulnerabilityObservationState(name="", port=-1, protocol=TransportProtocol.TCP, cvss=0.5,
                                                      osvdbid=-1, description="", credentials=[Credential.schema()],
                                                      service="")



from csle_common.util.ssh_util import SSHUtil
from unittest.mock import patch, MagicMock
import pytest


class TestSSHUtilSuite:
    """
    Test suite for ssh_util
    """

    @pytest.fixture(autouse=True)
    def mock_sleep(self):
        """
        Mock time.sleep to avoid delays
        """
        with patch("time.sleep", return_value=None):
            yield

    @patch("csle_common.util.ssh_util.SSHUtil.execute_ssh_cmd")
    def test_execute_ssh_cmds(self, mock_execute_ssh_cmd) -> None:
        """
        Test the method that executes a list of commands over an ssh connection to the emulation

        :param mock_execute_ssh_cmd: mock_execute_ssh_cmd
        :return: None
        """
        mock_execute_ssh_cmd.return_value = (b"output", b"error", 1.0)
        cmds = ["ls", "pwd", "whoami"]
        conn = MagicMock()
        results = SSHUtil.execute_ssh_cmds(cmds, conn)
        mock_execute_ssh_cmd.assert_called()
        assert results == [(b"output", b"error", 1.0)] * len(cmds)

    def test_execute_ssh_cmd(self) -> None:
        """
        Test the method that executes an action on the emulation over a ssh connection

        :return: None
        """
        conn = MagicMock()
        mock_transport = MagicMock()
        mock_session = MagicMock()
        mock_session.exit_status_ready.return_value = True
        mock_session.recv_ready.return_value = True
        mock_session.recv_stderr_ready.return_value = True
        mock_session.recv.side_effect = [b"output", b""]
        mock_session.recv_stderr.side_effect = [b"error", b""]
        conn.get_transport.return_value = mock_transport
        mock_transport.open_session.return_value = mock_session

        with pytest.raises(ConnectionError, match="Connection failed"):
            SSHUtil.execute_ssh_cmd("ls", conn)



from typing import List
import threading
import time
from confluent_kafka import Consumer, KafkaError, KafkaException
import csle_collector.constants.constants as collector_constants
from csle_collector.docker_stats_manager.dao.docker_stats import DockerStats
from csle_common.logging.log import Logger


class DockerStatsConsumerThread(threading.Thread):
    """
    Thread that polls the Kafka log to get the latest status of the docker statistics
    """

    def __init__(self, kafka_server_ip: str, kafka_port: int, docker_stats: DockerStats,
                 auto_offset_reset: str = "latest") -> None:
        """
        Initializes the thread

        :param kafka_server_ip: the ip of the kafka server
        :param kafka_port: the port of the kafka server
        :param docker_stats: the docker stats to update
        :param auto_offset_reset: the kafka offset to start reading from
        """
        threading.Thread.__init__(self)
        self.running = True
        self.kafka_server_ip = kafka_server_ip
        self.kafka_port = kafka_port
        self.docker_stats = docker_stats
        self.docker_stats_list: List[DockerStats] = []
        self.auto_offset_reset = auto_offset_reset
        self.ts = time.time()
        self.kafka_conf = {
            collector_constants.KAFKA.BOOTSTRAP_SERVERS_PROPERTY: f"{self.kafka_server_ip}:{self.kafka_port}",
            collector_constants.KAFKA.GROUP_ID_PROPERTY: f"docker_stats_consumer_thread_{self.ts}",
            collector_constants.KAFKA.AUTO_OFFSET_RESET_PROPERTY: auto_offset_reset}
        self.consumer = Consumer(**self.kafka_conf)
        self.consumer.subscribe([collector_constants.KAFKA_CONFIG.DOCKER_STATS_TOPIC_NAME])

    def run(self) -> None:
        """
        Runs the thread

        :return: None
        """
        while self.running:
            msg = self.consumer.poll(timeout=5.0)
            if msg is not None:
                if msg.error():
                    if msg.error().code() == KafkaError._PARTITION_EOF:
                        Logger.__call__().get_logger.warning(
                            f"reached end of partition: {msg.topic(), msg.partition(), msg.offset()}")
                    elif msg.error():
                        raise KafkaException(msg.error())
                else:
                    self.docker_stats.update_with_kafka_record(record=msg.value().decode())
                    self.docker_stats_list.append(self.docker_stats.copy())
        self.consumer.close()

    def get_average_docker_stats(self) -> DockerStats:
        """
        :return: the average of the list of docker stats
        """
        if len(self.docker_stats_list) == 0:
            return self.docker_stats.copy()
        if len(self.docker_stats_list) == 1:
            return self.docker_stats_list[0].copy()
        avg_docker_stats = DockerStats()
        for i in range(len(self.docker_stats_list)):
            avg_docker_stats.pids = avg_docker_stats.pids + self.docker_stats_list[i].pids
            avg_docker_stats.cpu_percent = avg_docker_stats.cpu_percent + self.docker_stats_list[i].cpu_percent
            avg_docker_stats.mem_current = avg_docker_stats.mem_current + self.docker_stats_list[i].mem_current
            avg_docker_stats.mem_total = avg_docker_stats.mem_total + self.docker_stats_list[i].mem_total
            avg_docker_stats.mem_percent = avg_docker_stats.mem_percent + self.docker_stats_list[i].mem_percent
            avg_docker_stats.blk_read = avg_docker_stats.blk_read + self.docker_stats_list[i].blk_read
            avg_docker_stats.blk_write = avg_docker_stats.blk_write + self.docker_stats_list[i].blk_write
            avg_docker_stats.net_rx = avg_docker_stats.net_rx + self.docker_stats_list[i].net_rx
            avg_docker_stats.net_tx = avg_docker_stats.pids + self.docker_stats_list[i].net_tx

        avg_docker_stats.pids = int(round(avg_docker_stats.pids / len(self.docker_stats_list)))
        avg_docker_stats.cpu_percent = int(round(avg_docker_stats.cpu_percent / len(self.docker_stats_list)))
        avg_docker_stats.mem_current = int(round(avg_docker_stats.mem_current / len(self.docker_stats_list)))
        avg_docker_stats.mem_total = int(round(avg_docker_stats.mem_total / len(self.docker_stats_list)))
        avg_docker_stats.mem_percent = int(round(avg_docker_stats.mem_percent / len(self.docker_stats_list)))
        avg_docker_stats.blk_read = int(round(avg_docker_stats.blk_read / len(self.docker_stats_list)))
        avg_docker_stats.blk_write = int(round(avg_docker_stats.blk_write / len(self.docker_stats_list)))
        avg_docker_stats.net_rx = int(round(avg_docker_stats.net_rx / len(self.docker_stats_list)))
        avg_docker_stats.net_tx = int(round(avg_docker_stats.net_tx / len(self.docker_stats_list)))
        return avg_docker_stats



from typing import Optional, List
import csle_common.constants.constants as constants
from csle_common.dao.emulation_action.attacker.emulation_attacker_action import EmulationAttackerAction
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_type import EmulationAttackerActionType
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_id import EmulationAttackerActionId
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_outcome import EmulationAttackerActionOutcome


class EmulationAttackerNIKTOActions:
    """
    Class implementing NIKTO scan actions for the attacker in the emulation
    """

    @staticmethod
    def NIKTO_WEB_HOST_SCAN(index: int, ips: Optional[List[str]] = None) -> EmulationAttackerAction:
        """
        Action for running a nikto web scan on a given host

        :param index: index of the machine to apply the action to
        :param ips: ip of the machine to apply the action to
        :return: the action
        """
        if ips is None:
            ips = []
        id = EmulationAttackerActionId.NIKTO_WEB_HOST_SCAN
        cmd = ["no | sudo nikto " + constants.NIKTO.BASE_ARGS]
        return EmulationAttackerAction(id=id, name="Nikto Web Scan", cmds=cmd,
                                       type=EmulationAttackerActionType.RECON,
                                       descr="Nikto Web Scan",
                                       ips=ips, index=index,
                                       action_outcome=EmulationAttackerActionOutcome.INFORMATION_GATHERING)



from unittest.mock import patch
from csle_common.util.multiprocessing_util import NoDaemonProcess
from csle_common.util.multiprocessing_util import NoDaemonContext
from csle_common.util.multiprocessing_util import NestablePool


class TestMultiprocessingUtilSuite:
    """
    Test suite for multiprocessing util
    """

    def test_daemon(self) -> None:
        """
        Test the process with daemon property set to false

        :return: None
        """
        result = NoDaemonProcess(target=lambda: None).daemon
        assert not result

    def test_no_daemon_context(self) -> None:
        """
        Test the NoDaemonContext method

        :return: None
        """
        context = NoDaemonContext()
        process = context.Process(target=lambda: None)
        assert isinstance(process, NoDaemonProcess)
        assert not process.daemon

    @patch("multiprocessing.get_context")
    def test_nestable_pool_initialization(self, mock_get_context) -> None:
        """
        Test the method that initializes the pool

        :param mock_get_context: mock_get_context
        :return: None
        """
        mock_get_context.return_value = NoDaemonContext()
        pool = NestablePool()
        assert pool



import logging
from typing import Any, Tuple
import subprocess
import docker
import psutil
import os
import csle_common.constants.constants as constants
import sys
import shutil


class ManagementSystemController:
    """
    Controller managing monitoring tools
    """

    @staticmethod
    def read_pid_file(path: str) -> int:
        """
        Reads the PID from a pidfile

        :param path: the path to the file
        :return: the parsed pid, or -1 if the pidfile could not be read
        """
        if os.path.exists(path):
            pid = int(open(path, "r").read())
            return pid
        return -1

    @staticmethod
    def is_prometheus_running() -> bool:
        """
        Checks if prometheus is running on the host

        :return: True if it is running, false otherwise
        """
        pid = ManagementSystemController.read_pid_file(constants.COMMANDS.PROMETHEUS_PID_FILE)
        if pid == -1:
            return False
        cmd = (constants.COMMANDS.PS_AUX + constants.COMMANDS.SPACE_DELIM + constants.COMMANDS.PIPE_DELIM +
               constants.COMMANDS.SPACE_DELIM + constants.COMMANDS.GREP
               + constants.COMMANDS.SPACE_DELIM + "prometheus")
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
        (output, err) = p.communicate()
        output_1 = str(output)
        return constants.COMMANDS.SEARCH_PROMETHEUS in output_1 and str(pid) in output_1

    @staticmethod
    def is_node_exporter_running() -> bool:
        """
        Checks if node_exporter is running on the host

        :return: True if it is running, false otherwise
        """
        pid = ManagementSystemController.read_pid_file(constants.COMMANDS.NODE_EXPORTER_PID_FILE)
        if pid == -1:
            return False
        cmd = (constants.COMMANDS.PS_AUX + constants.COMMANDS.SPACE_DELIM + constants.COMMANDS.PIPE_DELIM +
               constants.COMMANDS.SPACE_DELIM + constants.COMMANDS.GREP + constants.COMMANDS.SPACE_DELIM +
               constants.COMMANDS.SEARCH_NODE_EXPORTER)
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
        (output, err) = p.communicate()
        output_1 = str(output)
        return constants.COMMANDS.SEARCH_NODE_EXPORTER in output_1 and str(pid) in output_1

    @staticmethod
    def is_nginx_running() -> bool:
        """
        Checks if Nginx is running on the host

        :return: True if nginx is running, false otherwise
        """
        output = ""
        try:
            output = subprocess.run(constants.COMMANDS.NGINX_STATUS.split(" "), capture_output=True, text=True).stdout
        except Exception:
            pass
        nginx_running = "active (running)" in output or "active (exited)" in output
        return nginx_running

    @staticmethod
    def is_postgresql_running() -> bool:
        """
        Checks if PostgreSQL is running on the host

        :return: True if PostgreSQL is running, false otherwise
        """
        output = ""
        try:
            output = subprocess.run(constants.COMMANDS.POSTGRESQL_STATUS.split(" "), capture_output=True,
                                    text=True).stdout
        except Exception:
            pass
        postgresql_running = "active (running)" in output or "active (exited)" in output
        if postgresql_running:
            return postgresql_running
        output = ""
        try:
            output = subprocess.run(constants.COMMANDS.POSTGRESQL_STATUS_VERSION.split(" "), capture_output=True,
                                    text=True).stdout
        except Exception:
            pass
        postgresql_running = "active (running)" in output or "active (exited)" in output
        return postgresql_running

    @staticmethod
    def is_docker_engine_running() -> bool:
        """
        Checks if Docker engine is running on the host

        :return: True if Docker engine is running, false otherwise
        """
        output = ""
        try:
            output = subprocess.run(constants.COMMANDS.DOCKER_ENGINE_STATUS.split(" "), capture_output=True,
                                    text=True).stdout
        except Exception:
            pass
        docker_engine_running = "active (running)" in output or "active (exited)" in output
        return docker_engine_running

    @staticmethod
    def is_flask_running() -> bool:
        """
        Checks if the flask web server is running on the host

        :return: True if it is running, false otherwise
        """
        pid = ManagementSystemController.read_pid_file(constants.COMMANDS.CSLE_MGMT_WEBAPP_PID_FILE)
        if pid == -1:
            return False
        cmd = (constants.COMMANDS.PS_AUX + constants.COMMANDS.SPACE_DELIM + constants.COMMANDS.PIPE_DELIM +
               constants.COMMANDS.SPACE_DELIM + constants.COMMANDS.GREP + constants.COMMANDS.SPACE_DELIM +
               constants.COMMANDS.SEARCH_MONITOR)
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
        (output, err) = p.communicate()
        output_1 = str(output)
        return constants.COMMANDS.SEARCH_MONITOR in output_1 and str(pid) in output_1

    @staticmethod
    def start_node_exporter(logger: logging.Logger) -> bool:
        """
        Starts the node exporter

        :param logger: the logger to use for logging
        :return: True if it was started, False otherwise
        """
        if ManagementSystemController.is_node_exporter_running():
            logger.info("Node exporter is already running")
            return False
        cmd = constants.COMMANDS.START_NODE_EXPORTER
        logger.info(f"Starting node exporter by running the command: {cmd}")
        p = subprocess.Popen(cmd, stdout=subprocess.DEVNULL, shell=True)
        p.communicate()
        return True

    @staticmethod
    def start_flask(logger: logging.Logger) -> bool:
        """
        Starts the Flask REST API Server

        :param logger: the logger to use for logging
        :return: True if it was started, False otherwise
        """
        if ManagementSystemController.is_flask_running():
            logger.info("Flask is already running")
            return False

        cmd = constants.COMMANDS.BUILD_CSLE_MGMT_WEBAPP
        logger.info(f"Building the web app with the command: {cmd}")
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
        while True:
            if p.stdout is None:
                raise ValueError("Cannot read due to NoneType")
            out = p.stdout.read(1)
            if p.poll() is not None:
                break
            if str(out) != '':
                try:
                    sys.stdout.write(out.decode("utf-8"))
                except Exception:
                    pass
                sys.stdout.flush()
        cmd = constants.COMMANDS.START_CSLE_MGMT_WEBAPP
        cmd = cmd.replace("python", str(shutil.which("python")))
        cmd = cmd.replace(f"${constants.CONFIG_FILE.CSLE_HOME_ENV_PARAM}",
                          os.environ[constants.CONFIG_FILE.CSLE_HOME_ENV_PARAM])
        logger.info(f"Starting flask with the command: {cmd}")
        p = subprocess.Popen(cmd.split(" "), stdout=subprocess.DEVNULL, shell=False)
        pid = p.pid

        cmd = constants.COMMANDS.SAVE_PID.format(pid, constants.COMMANDS.CSLE_MGMT_WEBAPP_PID_FILE)
        p = subprocess.Popen(cmd, stdout=subprocess.DEVNULL, shell=True)
        p.communicate()
        return True

    @staticmethod
    def stop_node_exporter(logger: logging.Logger) -> bool:
        """
        Stops the node exporter

        :param logger: the logger to use for logging
        :return: True if it was stopped, False otherwise
        """
        if not ManagementSystemController.is_node_exporter_running():
            logger.info("Node exporter is not running")
            return False
        pid = ManagementSystemController.read_pid_file(constants.COMMANDS.NODE_EXPORTER_PID_FILE)
        cmd = constants.COMMANDS.KILL_PROCESS.format(pid)
        logger.info(f"Stopping node exporter by running the command: {cmd}")
        p = subprocess.Popen(cmd, stdout=subprocess.DEVNULL, shell=True)
        p.communicate()
        return True

    @staticmethod
    def stop_flask(logger: logging.Logger) -> bool:
        """
        Stops the flask REST API

        :param logger: the logger to use for logging
        :return: True if it was stopped, False otherwise
        """
        if not ManagementSystemController.is_flask_running():
            logger.info("Flask is not running")
            return False
        pid = ManagementSystemController.read_pid_file(constants.COMMANDS.CSLE_MGMT_WEBAPP_PID_FILE)
        cmd = constants.COMMANDS.KILL_PROCESS.format(pid)
        logger.info(f"Stopping flask by running the command: {cmd}")
        p = subprocess.Popen(cmd, stdout=subprocess.DEVNULL, shell=True)
        p.communicate()
        return True

    @staticmethod
    def start_prometheus(logger: logging.Logger) -> bool:
        """
        Starts Prometheus

        :param logger: the logger to use for logging
        :return: True if it was started, False otherwise
        """
        if ManagementSystemController.is_prometheus_running():
            logger.info("Prometheus is already running")
            return False
        cmd = constants.COMMANDS.START_PROMETHEUS
        logger.info(f"Starting Prometheus by running the command: {cmd}")
        p = subprocess.Popen(cmd, stdout=subprocess.DEVNULL, shell=True)
        p.communicate()
        return True

    @staticmethod
    def stop_prometheus(logger: logging.RootLogger) -> bool:
        """
        Stops Prometheus

        :param logger: the logger to use for logging
        :return: True if it was stopped, False otherwise
        """
        if not ManagementSystemController.is_prometheus_running():
            logger.info("Prometheus is not running")
            return False
        pid = ManagementSystemController.read_pid_file(constants.COMMANDS.PROMETHEUS_PID_FILE)
        cmd = constants.COMMANDS.KILL_PROCESS.format(pid)
        logger.info(f"Stopping Prometheus by running the command: {cmd}")
        p = subprocess.Popen(cmd, stdout=subprocess.DEVNULL, shell=True)
        p.communicate()
        return True

    @staticmethod
    def is_cadvisor_running() -> bool:
        """
        :return: True if cadvisor is running, otherwise False
        """
        client_1 = docker.from_env()
        containers = client_1.containers.list()
        for c in containers:
            if constants.CONTAINER_IMAGES.CADVISOR in c.name:
                return True
        return False

    @staticmethod
    def is_pgadmin_running() -> bool:
        """
        :return: True if pgadmin is running, otherwise False
        """
        client_1 = docker.from_env()
        containers = client_1.containers.list()
        for c in containers:
            if constants.CONTAINER_IMAGES.PGADMIN in c.name:
                return True
        return False

    @staticmethod
    def stop_cadvisor() -> bool:
        """
        :return: True if cadvisor was stopped, otherwise False
        """
        client_1 = docker.from_env()
        containers = client_1.containers.list()
        for c in containers:
            if constants.CONTAINER_IMAGES.CADVISOR in c.name:
                c.stop()
                return True
        return False

    @staticmethod
    def stop_pgadmin() -> bool:
        """
        :return: True if pgadmin was stopped, otherwise False
        """
        client_1 = docker.from_env()
        containers = client_1.containers.list()
        for c in containers:
            if constants.CONTAINER_IMAGES.PGADMIN in c.name:
                c.stop()
                return True
        return False

    @staticmethod
    def start_cadvisor(logger: logging.Logger) -> bool:
        """
        Starts cAdvisor

        :param logger: the logger to use for logging
        :return: True if cadvisor was started, otherwise False
        """
        if ManagementSystemController.is_cadvisor_running():
            logger.info("cAdvisor is already running")
            return False
        client_1 = docker.from_env()
        containers = client_1.containers.list(all=True)
        for c in containers:
            if constants.CONTAINER_IMAGES.CADVISOR in c.name:
                container_state = c.attrs['State']
                running = container_state['Status'] == "running"
                if not running:
                    c.start()
                    return True
        cmd = constants.COMMANDS.START_CADVISOR
        logger.info(f"Starting cAdvisor by running the command: {cmd}")
        p = subprocess.Popen(cmd, stdout=subprocess.DEVNULL, shell=True)
        (output, err) = p.communicate()
        return True

    @staticmethod
    def start_postgresql(logger: logging.RootLogger) -> Tuple[bool, Any, Any]:
        """
        Starts PostgreSQL

        :param logger: the logger to use for logging
        :return: True if postgresql was started, otherwise False
        """
        if ManagementSystemController.is_postgresql_running():
            logger.info("PostgreSQL is already running")
            return False, None, None
        cmd = constants.COMMANDS.POSTGRESQL_START
        logger.info(f"Starting PostgreSQL by running the command: {cmd}")
        output = subprocess.run(cmd.split(" "), capture_output=True, text=True)
        return True, output.stdout, output.stderr

    @staticmethod
    def stop_postgresql(logger: logging.RootLogger) -> Tuple[bool, Any, Any]:
        """
        Stops PostgreSQL

        :param logger: the logger to use for logging
        :return: True if postgresql was stopped, otherwise False
        """
        if not ManagementSystemController.is_postgresql_running():
            logger.info("PostgreSQL is not running")
            return False, None, None
        cmd = constants.COMMANDS.POSTGRESQL_STOP
        logger.info(f"Stopping PostgreSQL by running the command: {cmd}")
        output = subprocess.run(constants.COMMANDS.POSTGRESQL_STOP.split(" "), capture_output=True, text=True)
        return True, output.stdout, output.stderr

    @staticmethod
    def start_nginx(logger: logging.RootLogger) -> Tuple[bool, Any, Any]:
        """
        Starts Nginx

        :param logger: the logger to use for logging
        :return: True if nginx was started, otherwise False
        """
        if ManagementSystemController.is_nginx_running():
            logger.info("Nginx is already running")
            return False, None, None
        cmd = constants.COMMANDS.NGINX_START
        logger.info(f"Starting Nginx by running the command: {cmd}")
        output = subprocess.run(cmd.split(" "), capture_output=True, text=True)
        return True, output.stdout, output.stderr

    @staticmethod
    def stop_nginx(logger: logging.RootLogger) -> Tuple[bool, Any, Any]:
        """
        Stops Nginx

        :param logger: the logger to use for logging
        :return: True if postgresql was stopped, otherwise False
        """
        if not ManagementSystemController.is_nginx_running():
            logger.info("Nginx is not running")
            return False, None, None
        cmd = constants.COMMANDS.NGINX_STOP
        logger.info(f"Stopping Nginx by running the command: {cmd}")
        output = subprocess.run(cmd.split(" "), capture_output=True, text=True)
        return True, output.stdout, output.stderr

    @staticmethod
    def start_docker_engine(logger: logging.RootLogger) -> Tuple[bool, Any, Any]:
        """
        Starts the Docker engine

        :param logger: the logger to use for logging
        :return: True if nginx was started, otherwise False
        """
        if ManagementSystemController.is_docker_engine_running():
            logger.info("The Docker engine is already running")
            return False, None, None
        cmd = constants.COMMANDS.DOCKER_ENGINE_START
        logger.info(f"Starting the Docker engine by running the command: {cmd}")
        output = subprocess.run(cmd.split(" "), capture_output=True, text=True)
        return True, output.stdout, output.stderr

    @staticmethod
    def stop_docker_engine(logger: logging.RootLogger) -> Tuple[bool, Any, Any]:
        """
        Stops the Docker engine

        :param logger: the logger to use for logging
        :return: True if postgresql was stopped, otherwise False
        """
        if not ManagementSystemController.is_docker_engine_running():
            logger.info("The Docker engine is not running")
            return False, None, None
        cmd = constants.COMMANDS.DOCKER_ENGINE_STOP
        logger.info(f"Stopping the Docker engine by running the command: {cmd}")
        output = subprocess.run(cmd.split(" "), capture_output=True, text=True)
        return True, output.stdout, output.stderr

    @staticmethod
    def start_pgadmin(logger: logging.RootLogger) -> bool:
        """
        Starts pgAdmin

        :param logger: the logger to use for logging
        :return: True if pgadmin was started, otherwise False
        """
        if ManagementSystemController.is_pgadmin_running():
            logger.info("pgAdmin is already running")
            return False
        client_1 = docker.from_env()
        containers = client_1.containers.list(all=True)
        for c in containers:
            if constants.CONTAINER_IMAGES.PGADMIN in c.name:
                container_state = c.attrs['State']
                running = container_state['Status'] == "running"
                if not running:
                    c.start()
                    return True
        cmd = constants.COMMANDS.START_PGADMIN
        logger.info(f"Starting pgAdmin by running the command: {cmd}")
        p = subprocess.Popen(cmd, stdout=subprocess.DEVNULL, shell=True)
        (output, err) = p.communicate()
        return True

    @staticmethod
    def stop_grafana() -> bool:
        """
        Stops Grafana

        :return: True if grafana was stopped, otherwise False
        """
        client_1 = docker.from_env()
        containers = client_1.containers.list()
        for c in containers:
            if constants.CONTAINER_IMAGES.GRAFANA in c.name:
                c.stop()
                return True
        return False

    @staticmethod
    def start_grafana(logger: logging.RootLogger) -> bool:
        """
        Starts Grafana

        :param logger: the logger to use for logging
        :return: True if grafana was started, otherwise False
        """
        if ManagementSystemController.is_grafana_running():
            logger.info("Grafana is already running")
            return False
        client_1 = docker.from_env()
        containers = client_1.containers.list(all=True)
        for c in containers:
            if constants.CONTAINER_IMAGES.GRAFANA in c.name:
                container_state = c.attrs['State']
                running = container_state['Status'] == "running"
                if not running:
                    c.start()
                    return True
        cmd = constants.COMMANDS.START_GRAFANA
        logger.info(f"Starting Grafana by running the command: {cmd}")
        p = subprocess.Popen(cmd, stdout=subprocess.DEVNULL, shell=True)
        (output, err) = p.communicate()
        return True

    @staticmethod
    def is_grafana_running() -> bool:
        """
        :return: True if grafana is running otherwise False
        """
        client_1 = docker.from_env()
        containers = client_1.containers.list()
        for c in containers:
            if constants.CONTAINER_IMAGES.GRAFANA in c.name:
                return True
        return False

    @staticmethod
    def is_statsmanager_running() -> bool:
        """
        Checks if the statsmanager is running on the host

        :return: True if it is running, false otherwise
        """
        pid = ManagementSystemController.read_pid_file(constants.COMMANDS.DOCKER_STATS_MANAGER_PIDFILE)
        if pid == -1:
            return False
        logging.getLogger().info(pid)
        cmd = (constants.COMMANDS.PS_AUX + constants.COMMANDS.SPACE_DELIM + constants.COMMANDS.PIPE_DELIM +
               constants.COMMANDS.SPACE_DELIM + constants.COMMANDS.GREP + constants.COMMANDS.SPACE_DELIM +
               constants.COMMANDS.SEARCH_DOCKER_STATS_MANAGER)
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
        (output, err) = p.communicate()
        output_1 = str(output)
        return constants.COMMANDS.SEARCH_DOCKER_STATS_MANAGER in output_1 and str(pid) in output_1

    @staticmethod
    def start_docker_statsmanager(logger: logging.Logger, log_file: str = "docker_stats_manager.log",
                                  log_dir: str = "/var/log/csle", max_workers: int = 10,
                                  port: int = 50046) -> bool:
        """
        Starts the docker stats manager on the docker host if it is not already started

        :param logger: the logger to use for logging
        :param port: the port that the docker stats manager will listen to
        :param log_file: log file of the docker stats manager
        :param log_dir: log dir of the docker stats manager
        :param max_workers: max workers of the docker stats manager
        :return: True if it was started, False otherwise
        """
        if ManagementSystemController.is_statsmanager_running():
            logger.info("The docker statsmanager is already running")
            return False
        cmd = constants.COMMANDS.START_DOCKER_STATS_MANAGER.format(port, log_dir, log_file, max_workers)
        logger.info(f"Starting the Docker stats manager by running the command: {cmd}")
        p = subprocess.Popen(cmd, stdout=subprocess.DEVNULL, shell=True)
        (output, err) = p.communicate()
        return True

    @staticmethod
    def stop_docker_statsmanager(logger: logging.Logger) -> bool:
        """
        Stops the statsmanager

        :param logger: the logger to use for logging
        :return: True if it was stopped, False otherwise
        """
        if not ManagementSystemController.is_statsmanager_running():
            logger.info("The statsmanager is not running")
            return False
        pid = ManagementSystemController.read_pid_file(constants.COMMANDS.DOCKER_STATS_MANAGER_PIDFILE)
        cmd = constants.COMMANDS.KILL_PROCESS.format(pid)
        logger.info(f"Stopping the statsmanager by running the command: {cmd}")
        p = subprocess.Popen(cmd, stdout=subprocess.DEVNULL, shell=True)
        (output, err) = p.communicate()
        return True

    @staticmethod
    def stop_cluster_manager() -> bool:
        """
        Stops the local cluster manager

        :return: True if it was stopped, False otherwise
        """
        pid = ManagementSystemController.read_pid_file(constants.COMMANDS.CLUSTER_MANAGER_PIDFILE)
        cmd = constants.COMMANDS.KILL_PROCESS.format(pid)
        p = subprocess.Popen(cmd, stdout=subprocess.DEVNULL, shell=True)
        (output, err) = p.communicate()
        return True

    @staticmethod
    def is_pid_running(pid: int, logger: logging.Logger) -> bool:
        """
        Checks if the given pid is running on the host

        :param pid: the pid to check
        :param logger: the logger to use for logging
        :return: True if it is running, false otherwise
        """
        logger.info(f"Checking if PID: {pid} is running")
        return bool(psutil.pid_exists(pid))

    @staticmethod
    def stop_pid(pid, logger: logging.Logger) -> bool:
        """
        Stops a process with a given pid

        :param pid: the pid to stop
        :return: True if the pid was stopped, false if it was not running
        """
        if not ManagementSystemController.is_pid_running(pid, logger=logger):
            return False
        cmd = constants.COMMANDS.KILL_PROCESS.format(pid)
        logger.info(f"Stopping PID:{pid} with command: {cmd}")
        p = subprocess.Popen(cmd, stdout=subprocess.DEVNULL, shell=True)
        p.communicate()
        return True



import logging
import subprocess
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.util.emulation_util import EmulationUtil


class ResourceConstraintsController:
    """
    Class managing resource constraints in the emulation environments
    """

    @staticmethod
    def apply_resource_constraints(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                                   logger: logging.Logger) -> None:
        """
        Creates users in an emulation environment according to a specified users-configuration

        :param emulation_env_config: the emulation env configuration
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        resource_constraints = emulation_env_config.resources_config.node_resources_configurations
        resource_constraints = resource_constraints + [emulation_env_config.kafka_config.resources]
        if emulation_env_config.sdn_controller_config is not None:
            resource_constraints = resource_constraints + [emulation_env_config.sdn_controller_config.resources]
        for node_resource_config in resource_constraints:
            if node_resource_config.physical_host_ip != physical_server_ip:
                continue
            ip = node_resource_config.docker_gw_bridge_ip
            logger.info(f"applying resource constraints on node: {ip}, "
                        f"{node_resource_config.container_name}")
            EmulationUtil.connect_admin(emulation_env_config=emulation_env_config, ip=ip)

            # update cpus and memory
            cmd = f"docker update --memory={node_resource_config.available_memory_gb}G " \
                  f"--cpus={node_resource_config.num_cpus} {node_resource_config.container_name}"
            subprocess.Popen(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, shell=True)

            for ip_and_net_config in node_resource_config.ips_and_network_configs:
                _, net_config = ip_and_net_config
                if net_config is not None:
                    # delete existing netem rules
                    cmd = f"sudo tc qdisc del dev {net_config.interface} root netem"
                    o, e, _ = EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip))

                    # add new netem rule that implements the traffic shaping configuration
                    cmd = f"sudo tc qdisc add dev {net_config.interface} root netem " \
                          f"delay {net_config.packet_delay_ms:.6f}ms " \
                          f"{net_config.packet_delay_jitter_ms:.6f}ms " \
                          f"distribution {str(net_config.packet_delay_distribution)} " \
                          f"loss gemodel {net_config.loss_gemodel_p:.6f} " \
                          f"{net_config.loss_gemodel_r:.6f} " \
                          f"{(1 - net_config.loss_gemodel_h):.6f} " \
                          f"{(1 - net_config.loss_gemodel_k):.6f} " \
                          f"duplicate {net_config.packet_duplicate_percentage:.6f}% " \
                          f"{net_config.packet_duplicate_correlation_percentage:.6f}% corrupt " \
                          f"{net_config.packet_corrupt_percentage:.6f}% " \
                          f"reorder {net_config.packet_reorder_percentage:.6f}% " \
                          f"{net_config.packet_reorder_correlation_percentage:.6f}% " \
                          f"gap {net_config.packet_reorder_gap} " \
                          f"rate {net_config.rate_limit_mbit:.6f}mbit " \
                          f"limit {net_config.limit_packets_queue:.6f}"
                    o, e, _ = EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip))
            EmulationUtil.disconnect_admin(emulation_env_config=emulation_env_config)



from typing import List, Dict, Any
from csle_collector.snort_ids_manager.dao.snort_ids_alert_counters import SnortIdsAlertCounters
from csle_collector.snort_ids_manager.dao.snort_ids_ip_alert_counters import SnortIdsIPAlertCounters
from csle_collector.snort_ids_manager.dao.snort_ids_rule_counters import SnortIdsRuleCounters
from csle_collector.ossec_ids_manager.dao.ossec_ids_alert_counters import OSSECIdsAlertCounters
from csle_collector.client_manager.client_population_metrics import ClientPopulationMetrics
from csle_collector.five_g_core_manager.dao.five_g_core_amf_metrics import FiveGCoreAMFMetrics
from csle_collector.five_g_core_manager.dao.five_g_core_hss_metrics import FiveGCoreHSSMetrics
from csle_collector.five_g_core_manager.dao.five_g_core_mme_metrics import FiveGCoreMMEMetrics
from csle_collector.five_g_core_manager.dao.five_g_core_upf_metrics import FiveGCoreUPFMetrics
from csle_collector.five_g_core_manager.dao.five_g_core_pcf_metrics import FiveGCorePCFMetrics
from csle_collector.five_g_core_manager.dao.five_g_core_pcrf_metrics import FiveGCorePCRFMetrics
from csle_collector.five_g_core_manager.dao.five_g_core_smf_metrics import FiveGCoreSMFMetrics
from csle_collector.docker_stats_manager.dao.docker_stats import DockerStats
from csle_collector.host_manager.dao.host_metrics import HostMetrics
from csle_ryu.dao.avg_port_statistic import AvgPortStatistic
from csle_ryu.dao.avg_flow_statistic import AvgFlowStatistic
from csle_ryu.dao.flow_statistic import FlowStatistic
from csle_ryu.dao.port_statistic import PortStatistic
from csle_ryu.dao.agg_flow_statistic import AggFlowStatistic
from csle_common.dao.emulation_action.attacker.emulation_attacker_action import EmulationAttackerAction
from csle_common.dao.emulation_action.defender.emulation_defender_action import EmulationDefenderAction
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_base.json_serializable import JSONSerializable


class EmulationMetricsTimeSeries(JSONSerializable):
    """
    DTO containing time series data from the emulation
    """

    def __init__(self, client_metrics: List[ClientPopulationMetrics], aggregated_docker_stats: List[DockerStats],
                 docker_host_stats: Dict[str, List[DockerStats]], host_metrics: Dict[str, List[HostMetrics]],
                 aggregated_host_metrics: List[HostMetrics],
                 defender_actions: List[EmulationDefenderAction], attacker_actions: List[EmulationAttackerAction],
                 agg_snort_ids_metrics: List[SnortIdsAlertCounters], emulation_env_config: EmulationEnvConfig,
                 ossec_host_alert_counters: Dict[str, List[OSSECIdsAlertCounters]],
                 aggregated_ossec_host_alert_counters: List[OSSECIdsAlertCounters],
                 openflow_flow_stats: List[FlowStatistic], openflow_port_stats: List[PortStatistic],
                 avg_openflow_flow_stats: List[AvgFlowStatistic], avg_openflow_port_stats: List[AvgPortStatistic],
                 openflow_flow_metrics_per_switch: Dict[str, List[FlowStatistic]],
                 openflow_port_metrics_per_switch: Dict[str, List[PortStatistic]],
                 openflow_flow_avg_metrics_per_switch: Dict[str, List[AvgFlowStatistic]],
                 openflow_port_avg_metrics_per_switch: Dict[str, List[AvgPortStatistic]],
                 agg_openflow_flow_metrics_per_switch: Dict[str, List[AggFlowStatistic]],
                 agg_openflow_flow_stats: List[AggFlowStatistic],
                 snort_ids_ip_metrics: Dict[str, List[SnortIdsIPAlertCounters]],
                 agg_snort_ids_rule_metrics: List[SnortIdsRuleCounters],
                 snort_alert_metrics_per_ids: Dict[str, List[SnortIdsAlertCounters]],
                 snort_rule_metrics_per_ids: Dict[str, List[SnortIdsRuleCounters]],
                 five_g_core_amf_metrics: Dict[str, List[FiveGCoreAMFMetrics]],
                 five_g_core_mme_metrics: Dict[str, List[FiveGCoreMMEMetrics]],
                 five_g_core_upf_metrics: Dict[str, List[FiveGCoreUPFMetrics]],
                 five_g_core_pcf_metrics: Dict[str, List[FiveGCorePCFMetrics]],
                 five_g_core_pcrf_metrics: Dict[str, List[FiveGCorePCRFMetrics]],
                 five_g_core_smf_metrics: Dict[str, List[FiveGCoreSMFMetrics]],
                 five_g_core_hss_metrics: Dict[str, List[FiveGCoreHSSMetrics]]
                 ):
        """
        Initializes the DTO

        :param client_metrics: Time series data with information about the client population
        :param aggregated_docker_stats: Time series data with average docker statistics
        :param docker_host_stats: Time series data with docker statistics per host
        :param host_metrics: Time series data with general host metrics
        :param aggregated_host_metrics: Time series data with aggregated host metrics
        :param defender_actions: Time series data with defender actions
        :param attacker_actions: Time series data with attacker actions
        :param agg_snort_ids_metrics: Time series data with Snort IDS metrics
        :param emulation_env_config: the emulation config
        :param ossec_host_alert_counters: Time series data with ossec alert counters per host
        :param aggregated_ossec_host_alert_counters: Time series data with aggregated ossec alert counters
        :param openflow_flow_stats: openflow flow statistics
        :param openflow_port_stats: openflow port statistics
        :param avg_openflow_flow_stats: average openflow flow statistics per switch
        :param avg_openflow_port_stats: average openflow port statistics per switch
        :param openflow_flow_metrics_per_switch: openflow flow statistics per aggregated per switch
        :param openflow_port_metrics_per_switch: openflow port statistics per aggregated per switch
        :param openflow_flow_avg_metrics_per_switch: average openflow flow statistics per aggregated per switch
        :param openflow_port_avg_metrics_per_switch: average openflow port statistics per aggregated per switch
        :param agg_openflow_flow_stats: aggregated openflow flow statistics
        :param agg_openflow_flow_metrics_per_switch: aggregated openflow flow statistics aggregatd per switch
        :param snort_ids_ip_metrics: Time series data with Snort IDS metrics per IP
        :param agg_snort_ids_rule_metrics: Time series data with Snort IDS metrics per rule
        :param snort_alert_metrics_per_ids: Time series data with Snort IDS alert metrics per IDS
        :param snort_rule_metrics_per_ids: Time series data with Snort IDS rule metrics per IDS
        :param five_g_core_amf_metrics: Time series data with FiveGCore AMF metrics per host
        :param five_g_core_mme_metrics: Time series data with FiveGCore MME metrics per host
        :param five_g_core_upf_metrics: Time series data with FiveGCore UPF metrics per host
        :param five_g_core_pcf_metrics: Time series data with FiveGCore PCF metrics per host
        :param five_g_core_pcrf_metrics: Time series data with FiveGCore PCRF metrics per host
        :param five_g_core_smf_metrics: Time series data with FiveGCore SMF metrics per host
        :param five_g_core_hss_metrics: Time series data with FiveGCore HSS metrics per host
        """
        self.client_metrics = client_metrics
        self.aggregated_docker_stats = aggregated_docker_stats
        self.docker_host_stats = docker_host_stats
        self.host_metrics = host_metrics
        self.defender_actions = defender_actions
        self.attacker_actions = attacker_actions
        self.agg_snort_ids_metrics = agg_snort_ids_metrics
        self.aggregated_host_metrics = aggregated_host_metrics
        self.emulation_env_config = emulation_env_config
        self.ossec_host_alert_counters = ossec_host_alert_counters
        self.aggregated_ossec_host_alert_counters = aggregated_ossec_host_alert_counters
        self.openflow_flow_stats = openflow_flow_stats
        self.openflow_port_stats = openflow_port_stats
        self.avg_openflow_flow_stats = avg_openflow_flow_stats
        self.avg_openflow_port_stats = avg_openflow_port_stats
        self.openflow_flow_metrics_per_switch = openflow_flow_metrics_per_switch
        self.openflow_port_metrics_per_switch = openflow_port_metrics_per_switch
        self.openflow_flow_avg_metrics_per_switch = openflow_flow_avg_metrics_per_switch
        self.openflow_port_avg_metrics_per_switch = openflow_port_avg_metrics_per_switch
        self.agg_openflow_flow_stats = agg_openflow_flow_stats
        self.agg_openflow_flow_metrics_per_switch = agg_openflow_flow_metrics_per_switch
        self.snort_ids_ip_metrics = snort_ids_ip_metrics
        self.agg_snort_ids_rule_metrics = agg_snort_ids_rule_metrics
        self.snort_alert_metrics_per_ids = snort_alert_metrics_per_ids
        self.snort_rule_metrics_per_ids = snort_rule_metrics_per_ids
        self.five_g_core_amf_metrics = five_g_core_amf_metrics
        self.five_g_core_mme_metrics = five_g_core_mme_metrics
        self.five_g_core_upf_metrics = five_g_core_upf_metrics
        self.five_g_core_pcf_metrics = five_g_core_pcf_metrics
        self.five_g_core_pcrf_metrics = five_g_core_pcrf_metrics
        self.five_g_core_smf_metrics = five_g_core_smf_metrics
        self.five_g_core_hss_metrics = five_g_core_hss_metrics

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "EmulationMetricsTimeSeries":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the converted instance
        """
        docker_host_stats = {}
        for k, v in d["docker_host_stats"].items():
            docker_host_stats[k] = list(map(lambda x: DockerStats.from_dict(x), v))
        host_metrics = {}
        for k, v in d["host_metrics"].items():
            host_metrics[k] = list(map(lambda x: HostMetrics.from_dict(x), v))
        ossec_host_alerts = {}
        for k, v in d["ossec_host_alert_counters"].items():
            ossec_host_alerts[k] = list(map(lambda x: OSSECIdsAlertCounters.from_dict(x), v))

        openflow_flow_metrics_per_switch = {}
        for k, v in d["openflow_flow_metrics_per_switch"].items():
            openflow_flow_metrics_per_switch[k] = list(map(lambda x: FlowStatistic.from_dict(x), v))

        openflow_port_metrics_per_switch = {}
        for k, v in d["openflow_port_metrics_per_switch"].items():
            openflow_port_metrics_per_switch[k] = list(map(lambda x: PortStatistic.from_dict(x), v))

        openflow_flow_avg_metrics_per_switch = {}
        for k, v in d["openflow_flow_avg_metrics_per_switch"].items():
            openflow_flow_avg_metrics_per_switch[k] = list(map(lambda x: AvgFlowStatistic.from_dict(x), v))

        openflow_port_avg_metrics_per_switch = {}
        for k, v in d["openflow_port_avg_metrics_per_switch"].items():
            openflow_port_avg_metrics_per_switch[k] = list(map(lambda x: AvgPortStatistic.from_dict(x), v))

        agg_openflow_flow_metrics_per_switch = {}
        for k, v in d["agg_openflow_flow_metrics_per_switch"].items():
            agg_openflow_flow_metrics_per_switch[k] = list(map(lambda x: AggFlowStatistic.from_dict(x), v))

        snort_ids_ip_metrics = {}
        for k, v in d["snort_ids_ip_metrics"].items():
            snort_ids_ip_metrics[k] = list(map(lambda x: SnortIdsIPAlertCounters.from_dict(x), v))

        snort_alert_metrics_per_ids = {}
        for k, v in d["snort_alert_metrics_per_ids"].items():
            snort_alert_metrics_per_ids[k] = list(map(lambda x: SnortIdsAlertCounters.from_dict(x), v))

        snort_rule_metrics_per_ids = {}
        for k, v in d["snort_rule_metrics_per_ids"].items():
            snort_rule_metrics_per_ids[k] = list(map(lambda x: SnortIdsRuleCounters.from_dict(x), v))

        five_g_core_amf_metrics = {}
        for k, v in d["five_g_core_amf_metrics"].items():
            five_g_core_amf_metrics[k] = list(map(lambda x: FiveGCoreAMFMetrics.from_dict(x), v))

        five_g_core_mme_metrics = {}
        for k, v in d["five_g_core_mme_metrics"].items():
            five_g_core_mme_metrics[k] = list(map(lambda x: FiveGCoreMMEMetrics.from_dict(x), v))

        five_g_core_upf_metrics = {}
        for k, v in d["five_g_core_upf_metrics"].items():
            five_g_core_upf_metrics[k] = list(map(lambda x: FiveGCoreUPFMetrics.from_dict(x), v))

        five_g_core_pcf_metrics = {}
        for k, v in d["five_g_core_pcf_metrics"].items():
            five_g_core_pcf_metrics[k] = list(map(lambda x: FiveGCorePCFMetrics.from_dict(x), v))

        five_g_core_pcrf_metrics = {}
        for k, v in d["five_g_core_pcrf_metrics"].items():
            five_g_core_pcrf_metrics[k] = list(map(lambda x: FiveGCorePCRFMetrics.from_dict(x), v))

        five_g_core_mme_metrics = {}
        for k, v in d["five_g_core_mme_metrics"].items():
            five_g_core_mme_metrics[k] = list(map(lambda x: FiveGCoreMMEMetrics.from_dict(x), v))

        five_g_core_smf_metrics = {}
        for k, v in d["five_g_core_smf_metrics"].items():
            five_g_core_smf_metrics[k] = list(map(lambda x: FiveGCoreSMFMetrics.from_dict(x), v))

        five_g_core_mme_metrics = {}
        for k, v in d["five_g_core_mme_metrics"].items():
            five_g_core_mme_metrics[k] = list(map(lambda x: FiveGCoreMMEMetrics.from_dict(x), v))

        five_g_core_hss_metrics = {}
        for k, v in d["five_g_core_hss_metrics"].items():
            five_g_core_hss_metrics[k] = list(map(lambda x: FiveGCoreHSSMetrics.from_dict(x), v))

        obj = EmulationMetricsTimeSeries(
            client_metrics=list(map(lambda x: ClientPopulationMetrics.from_dict(x), d["client_metrics"])),
            aggregated_docker_stats=list(map(lambda x: DockerStats.from_dict(x), d["aggregated_docker_stats"])),
            docker_host_stats=docker_host_stats,
            host_metrics=host_metrics,
            defender_actions=list(map(lambda x: EmulationDefenderAction.from_dict(x), d["defender_actions"])),
            attacker_actions=list(map(lambda x: EmulationAttackerAction.from_dict(x), d["attacker_actions"])),
            agg_snort_ids_metrics=list(map(lambda x: SnortIdsAlertCounters.from_dict(x), d["agg_snort_ids_metrics"])),
            aggregated_host_metrics=list(map(lambda x: HostMetrics.from_dict(x), d["aggregated_host_metrics"])),
            emulation_env_config=EmulationEnvConfig.from_dict(d["emulation_env_config"]),
            ossec_host_alert_counters=ossec_host_alerts,
            aggregated_ossec_host_alert_counters=list(map(lambda x: OSSECIdsAlertCounters.from_dict(x),
                                                          d["aggregated_ossec_host_alert_counters"])),
            openflow_flow_stats=list(map(lambda x: FlowStatistic.from_dict(x), d["openflow_flow_stats"])),
            openflow_port_stats=list(map(lambda x: PortStatistic.from_dict(x), d["openflow_port_stats"])),
            avg_openflow_flow_stats=list(map(lambda x: AvgFlowStatistic.from_dict(x), d["avg_openflow_flow_stats"])),
            avg_openflow_port_stats=list(map(lambda x: AvgPortStatistic.from_dict(x), d["avg_openflow_port_stats"])),
            openflow_flow_metrics_per_switch=openflow_flow_metrics_per_switch,
            openflow_port_metrics_per_switch=openflow_port_metrics_per_switch,
            openflow_flow_avg_metrics_per_switch=openflow_flow_avg_metrics_per_switch,
            openflow_port_avg_metrics_per_switch=openflow_port_avg_metrics_per_switch,
            agg_openflow_flow_stats=list(map(lambda x: AggFlowStatistic.from_dict(x), d["agg_openflow_flow_stats"])),
            agg_openflow_flow_metrics_per_switch=agg_openflow_flow_metrics_per_switch,
            snort_ids_ip_metrics=snort_ids_ip_metrics,
            agg_snort_ids_rule_metrics=list(map(lambda x: SnortIdsRuleCounters.from_dict(x),
                                                d["agg_snort_ids_rule_metrics"])),
            snort_alert_metrics_per_ids=snort_alert_metrics_per_ids,
            snort_rule_metrics_per_ids=snort_rule_metrics_per_ids,
            five_g_core_amf_metrics=five_g_core_amf_metrics, five_g_core_mme_metrics=five_g_core_mme_metrics,
            five_g_core_upf_metrics=five_g_core_upf_metrics, five_g_core_pcf_metrics=five_g_core_pcf_metrics,
            five_g_core_pcrf_metrics=five_g_core_pcrf_metrics, five_g_core_smf_metrics=five_g_core_smf_metrics,
            five_g_core_hss_metrics=five_g_core_hss_metrics)
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["client_metrics"] = list(map(lambda x: x.to_dict(), self.client_metrics))
        d["aggregated_docker_stats"] = list(map(lambda x: x.to_dict(), self.aggregated_docker_stats))
        d["docker_host_stats"] = {}
        for k, v in self.docker_host_stats.items():
            d["docker_host_stats"][k] = list(map(lambda x: x.to_dict(), v))
        d["host_metrics"] = {}
        for k, v in self.host_metrics.items():
            d["host_metrics"][k] = list(map(lambda x: x.to_dict(), v))
        d["defender_actions"] = list(map(lambda x: x.to_dict(), self.defender_actions))
        d["attacker_actions"] = list(map(lambda x: x.to_dict(), self.attacker_actions))
        d["agg_snort_ids_metrics"] = list(map(lambda x: x.to_dict(), self.agg_snort_ids_metrics))
        d["aggregated_host_metrics"] = list(map(lambda x: x.to_dict(), self.aggregated_host_metrics))
        d["emulation_env_config"] = self.emulation_env_config.to_dict()
        d["aggregated_ossec_host_alert_counters"] = list(map(lambda x: x.to_dict(),
                                                             self.aggregated_ossec_host_alert_counters))
        d["ossec_host_alert_counters"] = {}
        for k, v in self.ossec_host_alert_counters.items():
            d["ossec_host_alert_counters"][k] = list(map(lambda x: x.to_dict(), v))
        d["openflow_flow_stats"] = list(map(lambda x: x.to_dict(), self.openflow_flow_stats))
        d["openflow_port_stats"] = list(map(lambda x: x.to_dict(), self.openflow_port_stats))
        d["avg_openflow_flow_stats"] = list(map(lambda x: x.to_dict(), self.avg_openflow_flow_stats))
        d["avg_openflow_port_stats"] = list(map(lambda x: x.to_dict(), self.avg_openflow_port_stats))

        d["openflow_flow_metrics_per_switch"] = {}
        for k, v in self.openflow_flow_metrics_per_switch.items():
            d["openflow_flow_metrics_per_switch"][k] = list(map(lambda x: x.to_dict(), v))

        d["openflow_port_metrics_per_switch"] = {}
        for k, v in self.openflow_port_metrics_per_switch.items():
            d["openflow_port_metrics_per_switch"][k] = list(map(lambda x: x.to_dict(), v))

        d["openflow_flow_avg_metrics_per_switch"] = {}
        for k, v in self.openflow_flow_avg_metrics_per_switch.items():
            d["openflow_flow_avg_metrics_per_switch"][k] = list(map(lambda x: x.to_dict(), v))

        d["openflow_port_avg_metrics_per_switch"] = {}
        for k, v in self.openflow_port_avg_metrics_per_switch.items():
            d["openflow_port_avg_metrics_per_switch"][k] = list(map(lambda x: x.to_dict(), v))

        d["agg_openflow_flow_metrics_per_switch"] = {}
        for k, v in self.agg_openflow_flow_metrics_per_switch.items():
            d["agg_openflow_flow_metrics_per_switch"][k] = list(map(lambda x: x.to_dict(), v))

        d["agg_openflow_flow_stats"] = list(map(lambda x: x.to_dict(), self.agg_openflow_flow_stats))

        d["snort_ids_ip_metrics"] = {}
        for k, v in self.snort_ids_ip_metrics.items():
            d["snort_ids_ip_metrics"][k] = list(map(lambda x: x.to_dict(), v))

        d["agg_snort_ids_rule_metrics"] = list(map(lambda x: x.to_dict(), self.agg_snort_ids_rule_metrics))

        d["snort_alert_metrics_per_ids"] = {}
        for k, v in self.snort_alert_metrics_per_ids.items():
            d["snort_alert_metrics_per_ids"][k] = list(map(lambda x: x.to_dict(), v))

        d["snort_rule_metrics_per_ids"] = {}
        for k, v in self.snort_rule_metrics_per_ids.items():
            d["snort_rule_metrics_per_ids"][k] = list(map(lambda x: x.to_dict(), v))

        d["five_g_core_amf_metrics"] = {}
        for k, v in self.five_g_core_amf_metrics.items():
            d["five_g_core_amf_metrics"][k] = list(map(lambda x: x.to_dict(), v))

        d["five_g_core_mme_metrics"] = {}
        for k, v in self.five_g_core_mme_metrics.items():
            d["five_g_core_mme_metrics"][k] = list(map(lambda x: x.to_dict(), v))

        d["five_g_core_upf_metrics"] = {}
        for k, v in self.five_g_core_upf_metrics.items():
            d["five_g_core_upf_metrics"][k] = list(map(lambda x: x.to_dict(), v))

        d["five_g_core_pcf_metrics"] = {}
        for k, v in self.five_g_core_pcf_metrics.items():
            d["five_g_core_pcf_metrics"][k] = list(map(lambda x: x.to_dict(), v))

        d["five_g_core_pcrf_metrics"] = {}
        for k, v in self.five_g_core_pcrf_metrics.items():
            d["five_g_core_pcrf_metrics"][k] = list(map(lambda x: x.to_dict(), v))

        d["five_g_core_smf_metrics"] = {}
        for k, v in self.five_g_core_smf_metrics.items():
            d["five_g_core_smf_metrics"][k] = list(map(lambda x: x.to_dict(), v))

        d["five_g_core_hss_metrics"] = {}
        for k, v in self.five_g_core_hss_metrics.items():
            d["five_g_core_hss_metrics"][k] = list(map(lambda x: x.to_dict(), v))

        return d

    def __str__(self) -> str:
        """
        :return: a string representation
        """
        return f"client_metrics: {list(map(lambda x: str(x), self.client_metrics))}," \
               f"aggregated_docker_stats: {list(map(lambda x: str(x), self.aggregated_docker_stats))}," \
               f"docker_host_stats: {list(map(lambda x: str(x), self.docker_host_stats))}," \
               f"host_metrics: {list(map(lambda x: str(x), self.host_metrics))}," \
               f"defender_actions: {list(map(lambda x: str(x), self.defender_actions))}," \
               f"attacker_actions: {list(map(lambda x: str(x), self.attacker_actions))}," \
               f"agg_snort_ids_metrics: {list(map(lambda x: str(x), self.agg_snort_ids_metrics))}," \
               f"aggregated_host_metrics: {list(map(lambda x: str(x), self.aggregated_host_metrics))}," \
               f"config: {self.emulation_env_config}," \
               f"aggregated_ossec_host_alert_counters: {self.aggregated_ossec_host_alert_counters}," \
               f"ossec_host_alert_counters: {self.ossec_host_alert_counters}," \
               f"openflow_flow_stats: {self.openflow_flow_stats}, openflow_port_stats: {self.openflow_port_stats}," \
               f"avg_openflow_flow_stats: {self.avg_openflow_flow_stats}, " \
               f"avg_openflow_port_stats: {self.avg_openflow_port_stats}," \
               f"openflow_flow_metrics_per_switch: {self.openflow_flow_metrics_per_switch}," \
               f"openflow_port_metrics_per_switch: {self.openflow_port_metrics_per_switch}," \
               f"openflow_flow_avg_metrics_per_switch: {self.openflow_flow_avg_metrics_per_switch}," \
               f"openflow_port_avg_metrics_per_switch: {self.openflow_port_avg_metrics_per_switch}," \
               f"agg_openflow_flow_stats: {self.agg_openflow_flow_stats}," \
               f"agg_openflow_flow_metrics_per_switch: {self.agg_openflow_flow_metrics_per_switch}," \
               f"snort_ids_ip_metrics: {self.snort_ids_ip_metrics}," \
               f"agg_snort_ids_rule_metrics: {self.agg_snort_ids_rule_metrics}," \
               f"snort_alert_metrics_per_ids: {self.snort_alert_metrics_per_ids}," \
               f"snort_rule_metrics_per_ids: {self.snort_rule_metrics_per_ids}" \
               f"five_g_core_amf_metrics: {list(map(lambda x: str(x), self.five_g_core_amf_metrics))}," \
               f"five_g_core_mme_metrics: {list(map(lambda x: str(x), self.five_g_core_mme_metrics))}," \
               f"five_g_core_upf_metrics: {list(map(lambda x: str(x), self.five_g_core_upf_metrics))}," \
               f"five_g_core_pcf_metrics: {list(map(lambda x: str(x), self.five_g_core_pcf_metrics))}," \
               f"five_g_core_pcrf_metrics: {list(map(lambda x: str(x), self.five_g_core_pcrf_metrics))}," \
               f"five_g_core_smf_metrics: {list(map(lambda x: str(x), self.five_g_core_smf_metrics))}," \
               f"five_g_core_hss_metrics: {list(map(lambda x: str(x), self.five_g_core_hss_metrics))}," \
 \
            @ staticmethod

    def from_json_file(json_file_path: str) -> "EmulationMetricsTimeSeries":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return EmulationMetricsTimeSeries.from_dict(json.loads(json_str))



from typing import List, Dict, Any, Union
from abc import abstractmethod
from numpy.typing import NDArray
from csle_common.dao.training.agent_type import AgentType
from csle_common.dao.training.player_type import PlayerType
from csle_base.json_serializable import JSONSerializable


class Policy(JSONSerializable):
    """
    An abstract class representing a policy
    """

    def __init__(self, agent_type: AgentType, player_type: PlayerType) -> None:
        """
        Initializes the object

        :param agent_type: the type of the agent using the policy
        :param player_type: the type of player using the policy (e.g. defender or attacker)
        """
        self.agent_type = agent_type
        self.player_type = player_type

    @abstractmethod
    def action(self, o: Any, deterministic: bool) -> Union[int, List[int], float, NDArray[Any]]:
        """
        Calculates the next action

        :param o: the input observation
        :param deterministic: boolean flag indicating whether the action selection should be deterministic
        :return: the action
        """
        pass

    @abstractmethod
    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        pass

    @abstractmethod
    def stage_policy(self, o: Any) -> Union[List[List[float]], List[float]]:
        """
        Returns a stage policy (see Horak & Bosansky 2019)

        :param o: the observation for the stage
        :return: the stage policy
        """
        pass

    @staticmethod
    @abstractmethod
    def from_dict(d: Dict[str, Any]) -> "Policy":
        """
        Converts a dict representation of the object to an instance

        :param d: the dict representation to convert
        :return: the converted object
        """
        pass

    @abstractmethod
    def probability(self, o: Any, a: int) -> Union[int, float]:
        """
        Calculates the probability of a given action for a given observation

        :param o: the observation
        :param a: the action
        :return: the probability
        """
        pass

    @abstractmethod
    def copy(self) -> "Policy":
        """
        :return: a copy of the object
        """
        pass

    @staticmethod
    @abstractmethod
    def from_json_file(json_file_path: str) -> "Policy":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        pass



from typing import Dict, Any, List
from csle_common.dao.emulation_config.cluster_node import ClusterNode
from csle_base.json_serializable import JSONSerializable


class ClusterConfig(JSONSerializable):
    """
    A DTO Class representing the config of a CSLE cluster
    """

    def __init__(self, cluster_nodes: List[ClusterNode]):
        """
        Initializes the DTO

        :param cluster_nodes: the list of cluster nodes
        """
        self.cluster_nodes = cluster_nodes

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d = {}
        d["cluster_nodes"] = list(map(lambda x: x.to_dict(), self.cluster_nodes))
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "ClusterConfig":
        """
        Convert a dict representation to a DTO representation

        :return: a dto representation of the object
        """
        dto = ClusterConfig(cluster_nodes=list(map(lambda x: ClusterNode.from_dict(x), d["cluster_nodes"])))
        return dto

    def __str__(self) -> str:
        """
        :return: a string representation of the credential
        """
        return f"cluster_nodes: {list(map(lambda x: str(x), self.cluster_nodes))}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "ClusterConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return ClusterConfig.from_dict(json.loads(json_str))

    def copy(self) -> "ClusterConfig":
        """
        :return: a copy of the DTO
        """
        return ClusterConfig.from_dict(self.to_dict())

    @staticmethod
    def schema() -> "ClusterConfig":
        """
        :return: get the schema of the DTO
        """
        return ClusterConfig(cluster_nodes=[])



from typing import List, Dict, Any
import csle_collector.five_g_core_manager.five_g_core_manager_pb2_grpc
import csle_collector.five_g_core_manager.five_g_core_manager_pb2
import csle_collector.five_g_core_manager.five_g_core_manager_util as five_g_core_manager_util
from csle_base.json_serializable import JSONSerializable


class FiveGCoreManagersInfo(JSONSerializable):
    """
    DTO containing the status of the 5G core managers for a given emulation execution
    """

    def __init__(
            self, ips: List[str], ports: List[int], emulation_name: str, execution_id: int,
            five_g_core_managers_statuses: List[
                csle_collector.five_g_core_manager.five_g_core_manager_pb2.FiveGCoreStatusDTO],
            five_g_core_managers_running: List[bool], local_webui_port: int = 9999, physical_server_ip: str = ""):
        """
        Initializes the DTO

        :param five_g_core_managers_running: list of booleans that indicate whether the 5G core managers are running
        :param ips: the list of IPs of the running 5G core managers
        :param ports: the list of ports of the running 5G core managers
        :param emulation_name: the name of the corresponding emulation
        :param execution_id: the ID of the corresponding emulation execution
        :param five_g_core_managers_statuses: a list of statuses of the 5G core managers
        :param local_webui_port: the local port of the Web UI
        :param physical_server_ip: the IP of the physical server
        """
        self.five_g_core_managers_running = five_g_core_managers_running
        self.ips = ips
        self.ports = ports
        self.emulation_name = emulation_name
        self.execution_id = execution_id
        self.five_g_core_managers_statuses = five_g_core_managers_statuses
        self.local_webui_port = local_webui_port
        self.physical_server_ip = physical_server_ip

    def __str__(self):
        """
        :return: a string representation of the DTO
        """
        return f"five_g_core_managers_running: {self.five_g_core_managers_running}, " \
               f"ips: {list(map(lambda x: str(x), self.ips))}, " \
               f"emulation_name: {self.emulation_name}, " \
               f"execution_id: {self.execution_id}, " \
               f"five_g_core_managers_statuses: {list(map(lambda x: str(x), self.five_g_core_managers_statuses))}," \
               f" ports: {list(map(lambda x: str(x), self.ports))}, local_webui_port: {self.local_webui_port}, " \
               f" physical_server_ip: {self.physical_server_ip}"

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["five_g_core_managers_running"] = self.five_g_core_managers_running
        d["ips"] = self.ips
        d["ports"] = self.ports
        d["emulation_name"] = self.emulation_name
        d["execution_id"] = self.execution_id
        d["local_webui_port"] = self.local_webui_port
        d["physical_server_ip"] = self.physical_server_ip
        d["five_g_core_managers_statuses"] = list(map(
            lambda x: five_g_core_manager_util.FiveGCoreManagerUtil.five_g_core_status_dto_to_dict(x),
            self.five_g_core_managers_statuses))
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "FiveGCoreManagersInfo":
        """
        Convert a dict representation to a DTO representation

        :return: a dto representation of the object
        """
        dto = FiveGCoreManagersInfo(
            five_g_core_managers_running=d["five_g_core_managers_running"], ips=d["ips"], ports=d["ports"],
            emulation_name=d["emulation_name"], execution_id=d["execution_id"],
            five_g_core_managers_statuses=list(map(
                lambda x: five_g_core_manager_util.FiveGCoreManagerUtil.five_g_core_status_dto_from_dict(x),
                d["five_g_core_managers_statuses"])),
            local_webui_port=d["local_webui_port"], physical_server_ip=d["physical_server_ip"]
        )
        return dto

    @staticmethod
    def get_empty_dto() -> "FiveGCoreManagersInfo":
        """
        :return: an empty version of the DTO
        """
        return FiveGCoreManagersInfo(
            ips=[], ports=[], emulation_name="", execution_id=-1, five_g_core_managers_statuses=[],
            five_g_core_managers_running=[], local_webui_port=9999, physical_server_ip="")

    @staticmethod
    def from_json_file(json_file_path: str) -> "FiveGCoreManagersInfo":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return FiveGCoreManagersInfo.from_dict(json.loads(json_str))



"""
Type of defense action outcomes
"""
from enum import IntEnum


class EmulationDefenderActionOutcome(IntEnum):
    """
    Enum representing the different defense outcomes in the network.
    """
    GAME_END = 0
    CONTINUE = 1
    STATE_UPDATE = 2
    ADD_DEFENSIVE_MECHANISM = 3



import logging
from typing import List
import grpc
import time
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.dao.emulation_config.snort_managers_info import SnortIdsManagersInfo
import csle_common.constants.constants as constants
import csle_collector.constants.constants as csle_collector_constants
import csle_collector.snort_ids_manager.snort_ids_manager_pb2_grpc
import csle_collector.snort_ids_manager.snort_ids_manager_pb2
import csle_collector.snort_ids_manager.query_snort_ids_manager
import csle_collector.snort_ids_manager.snort_ids_manager_util
from csle_common.util.emulation_util import EmulationUtil


class SnortIDSController:
    """
    Class managing operations related to Snort IDS Managers
    """

    @staticmethod
    def start_snort_idses(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                          logger: logging.Logger) -> None:
        """
        Utility function for starting the Snort IDSes

        :param emulation_env_config: the emulation env configuration
        :param physical_server_ip: the ip of the phsyical server
        :param logger: the logger to use for logging
        :return: None
        """
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            for ids_image in constants.CONTAINER_IMAGES.SNORT_IDS_IMAGES:
                if ids_image in c.name:
                    logger.info(f"Starting the Snort IDS on IP: {c.docker_gw_bridge_ip} "
                                f"({c.get_ips()[0]}, {c.get_full_name()})")
                    SnortIDSController.start_snort_ids(emulation_env_config=emulation_env_config,
                                                       ip=c.docker_gw_bridge_ip, logger=logger)

    @staticmethod
    def stop_snort_idses(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                         logger: logging.Logger) -> None:
        """
        Utility function for stopping the Snort IDSes

        :param emulation_env_config: the emulation env configuration
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            for ids_image in constants.CONTAINER_IMAGES.SNORT_IDS_IMAGES:
                if ids_image in c.name:
                    SnortIDSController.stop_snort_ids(emulation_env_config=emulation_env_config,
                                                      ip=c.docker_gw_bridge_ip, logger=logger)

    @staticmethod
    def start_snort_ids(emulation_env_config: EmulationEnvConfig, ip: str, logger: logging.Logger) -> None:
        """
        Utility function for starting the Snort IDS on a specific IP

        :param emulation_env_config: the emulation env configuration
        :param ip: the ip of the container
        :param logger: the logger to use for logging
        :return: None
        """
        SnortIDSController.start_snort_manager(emulation_env_config=emulation_env_config, ip=ip, logger=logger)
        ids_monitor_dto = SnortIDSController.get_snort_idses_monitor_threads_statuses_by_ip_and_port(
            port=emulation_env_config.snort_ids_manager_config.snort_ids_manager_port, ip=ip)
        if not ids_monitor_dto.snort_ids_running:
            logger.info(f"Snort IDS is not running on {ip}, starting it.")
            subnetmask = f"{emulation_env_config.execution_id}.{emulation_env_config.level}" \
                         f"{constants.CSLE.CSLE_LEVEL_SUBNETMASK_SUFFIX}"
            ingress_interface = constants.NETWORKING.ETH2
            egress_interface = constants.NETWORKING.ETH0
            logger.info(f"Subnetmask: {subnetmask}, ingress interface: {ingress_interface}, "
                        f"egress interface: {egress_interface}")
            # Open a gRPC session
            with grpc.insecure_channel(
                    f'{ip}:'
                    f'{emulation_env_config.snort_ids_manager_config.snort_ids_manager_port}',
                    options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
                stub = csle_collector.snort_ids_manager.snort_ids_manager_pb2_grpc.SnortIdsManagerStub(channel)
                csle_collector.snort_ids_manager.query_snort_ids_manager.start_snort_ids(
                    stub=stub, ingress_interface=ingress_interface, egress_interface=egress_interface,
                    subnetmask=subnetmask)

    @staticmethod
    def stop_snort_ids(emulation_env_config: EmulationEnvConfig, ip: str, logger: logging.Logger) -> None:
        """
        Utility function for stopping the Snort IDS on a specific IP

        :param emulation_env_config: the emulation env configuration
        :param ip: the ip of the container
        :param logger: the logger to use for logging
        :return: None
        """
        SnortIDSController.start_snort_manager(emulation_env_config=emulation_env_config, ip=ip, logger=logger)
        ids_monitor_dto = SnortIDSController.get_snort_idses_monitor_threads_statuses_by_ip_and_port(
            port=emulation_env_config.snort_ids_manager_config.snort_ids_manager_port, ip=ip)
        if ids_monitor_dto.snort_ids_running:
            logger.info(f"Snort IDS is running on {ip}, stopping it.")
            # Open a gRPC session
            with grpc.insecure_channel(
                    f'{ip}:'
                    f'{emulation_env_config.snort_ids_manager_config.snort_ids_manager_port}',
                    options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
                stub = csle_collector.snort_ids_manager.snort_ids_manager_pb2_grpc.SnortIdsManagerStub(channel)
                csle_collector.snort_ids_manager.query_snort_ids_manager.stop_snort_ids(stub=stub)

    @staticmethod
    def start_snort_managers(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                             logger: logging.Logger) -> None:
        """
        Utility function for starting snort IDS managers

        :param emulation_env_config: the emulation env config
        :param physical_server_ip: the physical server ip
        :param logger: the logger to use for logging
        :return: None
        """
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            for ids_image in constants.CONTAINER_IMAGES.SNORT_IDS_IMAGES:
                if ids_image in c.name:
                    SnortIDSController.start_snort_manager(emulation_env_config=emulation_env_config,
                                                           ip=c.docker_gw_bridge_ip, logger=logger)

    @staticmethod
    def start_snort_manager(emulation_env_config: EmulationEnvConfig, ip: str, logger: logging.Logger) -> None:
        """
        Utility function for starting the snort IDS manager on a specific IP

        :param emulation_env_config: the emulation env config
        :param ip: IP of the container
        :param logger: the logger to use for logging
        :return: None
        """
        # Check if ids_manager is already running
        status = None
        not_running = False
        try:
            status = SnortIDSController.get_snort_idses_monitor_threads_statuses_by_ip_and_port(
                ip=ip,
                port=emulation_env_config.snort_ids_manager_config.snort_ids_manager_port, timeout=5)
        except Exception:
            not_running = True
        status_str = ""
        if status is None:
            not_running = True
        else:
            status_str = f"monitor running: {status.monitor_running}, snort_ids_running: {status.snort_ids_running}"

        if not_running:
            logger.info(f"Starting Snort IDS manager on node {ip}")

            # Connect
            EmulationUtil.connect_admin(emulation_env_config=emulation_env_config, ip=ip)

            # Stop old background job if running
            cmd = (constants.COMMANDS.SUDO + constants.COMMANDS.SPACE_DELIM + constants.COMMANDS.PKILL +
                   constants.COMMANDS.SPACE_DELIM + constants.TRAFFIC_COMMANDS.SNORT_IDS_MANAGER_FILE_NAME)
            o, e, _ = EmulationUtil.execute_ssh_cmd(
                cmd=cmd, conn=emulation_env_config.get_connection(ip=ip))

            # Start the ids_manager
            cmd = constants.COMMANDS.START_SNORT_IDS_MANAGER.format(
                emulation_env_config.snort_ids_manager_config.snort_ids_manager_port,
                emulation_env_config.snort_ids_manager_config.snort_ids_manager_log_dir,
                emulation_env_config.snort_ids_manager_config.snort_ids_manager_log_file,
                emulation_env_config.snort_ids_manager_config.snort_ids_manager_max_workers)
            o, e, _ = EmulationUtil.execute_ssh_cmd(
                cmd=cmd, conn=emulation_env_config.get_connection(ip=ip))
            time.sleep(2)
        else:
            logger.info(f"The Snort IDS manager was already running on node {ip}. Status: {status_str}")

    @staticmethod
    def stop_snort_managers(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                            logger: logging.Logger) -> None:
        """
        Utility function for stopping snort IDS managers

        :param emulation_env_config: the emulation env config
        :param physical_server_ip: the physical server ip
        :param logger: the logger to use for logging
        :return: None
        """
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            for ids_image in constants.CONTAINER_IMAGES.SNORT_IDS_IMAGES:
                if ids_image in c.name:
                    SnortIDSController.stop_snort_manager(emulation_env_config=emulation_env_config,
                                                          ip=c.docker_gw_bridge_ip, logger=logger)

    @staticmethod
    def stop_snort_manager(emulation_env_config: EmulationEnvConfig, ip: str, logger: logging.Logger) -> None:
        """
        Utility function for stopping the snort IDS manager on a specific IP

        :param emulation_env_config: the emulation env config
        :param ip: the IP of the container
        :param logger: the logger to use for logging
        :return: None
        """
        # Connect
        EmulationUtil.connect_admin(emulation_env_config=emulation_env_config, ip=ip)
        logger.info(f"Stopping Snort IDS manager on node {ip}")

        cmd = (constants.COMMANDS.SUDO + constants.COMMANDS.SPACE_DELIM + constants.COMMANDS.PKILL +
               constants.COMMANDS.SPACE_DELIM + constants.TRAFFIC_COMMANDS.SNORT_IDS_MANAGER_FILE_NAME)
        o, e, _ = EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip))

        time.sleep(2)

    @staticmethod
    def start_snort_idses_monitor_threads(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                                          logger: logging.Logger) -> None:
        """
        A method that sends a request to the SnortIDSManager on every container that runs
        an IDS to start the IDS manager and the monitor thread

        :param emulation_env_config: the emulation env config
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            for ids_image in constants.CONTAINER_IMAGES.SNORT_IDS_IMAGES:
                if ids_image in c.name:
                    logger.info(f"Starting Snort IDS monitor thread on IP: {c.docker_gw_bridge_ip} "
                                f"({c.get_ips()[0]}, {c.get_full_name()})")
                    SnortIDSController.start_snort_idses_monitor_thread(emulation_env_config=emulation_env_config,
                                                                        ip=c.docker_gw_bridge_ip, logger=logger)

    @staticmethod
    def start_snort_idses_monitor_thread(emulation_env_config: EmulationEnvConfig, ip: str,
                                         logger: logging.Logger) -> None:
        """
        A method that sends a request to the SnortIDSManager on a specific container that runs
        an IDS to start the IDS manager and the monitor thread

        :param emulation_env_config: the emulation env config
        :param ip: the ip of the container
        :param logger: the logger to use for logging
        :return: None
        """
        SnortIDSController.start_snort_manager(emulation_env_config=emulation_env_config, ip=ip, logger=logger)

        ids_monitor_dto = SnortIDSController.get_snort_idses_monitor_threads_statuses_by_ip_and_port(
            port=emulation_env_config.snort_ids_manager_config.snort_ids_manager_port, ip=ip)
        if not ids_monitor_dto.monitor_running:
            logger.info(f"Snort IDS monitor thread is not running on {ip}, starting it.")
            # Open a gRPC session
            with grpc.insecure_channel(
                    f'{ip}:'
                    f'{emulation_env_config.snort_ids_manager_config.snort_ids_manager_port}',
                    options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
                stub = csle_collector.snort_ids_manager.snort_ids_manager_pb2_grpc.SnortIdsManagerStub(channel)
                csle_collector.snort_ids_manager.query_snort_ids_manager.start_snort_ids_monitor(
                    stub=stub, kafka_ip=emulation_env_config.kafka_config.container.get_ips()[0],
                    kafka_port=emulation_env_config.kafka_config.kafka_port,
                    log_file_path=csle_collector_constants.SNORT_IDS_ROUTER.SNORT_FAST_LOG_FILE,
                    time_step_len_seconds=emulation_env_config.snort_ids_manager_config.time_step_len_seconds)

    @staticmethod
    def stop_snort_idses_monitor_threads(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                                         logger: logging.Logger) -> None:
        """
        A method that sends a request to the SnortIDSManager on every container that runs
        an IDS to stop the monitor threads

        :param emulation_env_config: the emulation env config
        :param physical_server_ip: the physical server ip
        :param logger: the logger to use for logging
        :return: None
        """
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            for ids_image in constants.CONTAINER_IMAGES.SNORT_IDS_IMAGES:
                if ids_image in c.name:
                    SnortIDSController.stop_snort_idses_monitor_thread(emulation_env_config=emulation_env_config,
                                                                       ip=c.docker_gw_bridge_ip, logger=logger)

    @staticmethod
    def stop_snort_idses_monitor_thread(emulation_env_config: EmulationEnvConfig, ip: str,
                                        logger: logging.Logger) -> None:
        """
        A method that sends a request to the SnortIDSManager on a specific container that runs
        an IDS to stop the monitor threads

        :param emulation_env_config: the emulation env config
        :param ip: the IP of the container
        :param logger: the logger to use for logging
        :return: None
        """
        SnortIDSController.start_snort_manager(emulation_env_config=emulation_env_config, ip=ip, logger=logger)
        # Open a gRPC session
        with grpc.insecure_channel(
                f'{ip}:'
                f'{emulation_env_config.snort_ids_manager_config.snort_ids_manager_port}',
                options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.snort_ids_manager.snort_ids_manager_pb2_grpc.SnortIdsManagerStub(channel)
            logger.info(f"Stopping the Snort IDS monitor thread on {ip}.")
            csle_collector.snort_ids_manager.query_snort_ids_manager.stop_snort_ids_monitor(stub=stub)

    @staticmethod
    def get_snort_idses_monitor_threads_statuses(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                                                 logger: logging.Logger, start_if_stopped: bool = True) -> \
            List[csle_collector.snort_ids_manager.snort_ids_manager_pb2.SnortIdsMonitorDTO]:
        """
        A method that sends a request to the SnortIDSManager on every container to get the status of the
        IDS monitor thread

        :param emulation_env_config: the emulation config
        :param start_if_stopped: whether to start the IDS monitor if it is stopped
        :param physical_server_ip: the physical server IP
        :param logger: the logger to use for logging
        :return: List of monitor thread statuses
        """
        statuses = []
        if start_if_stopped:
            SnortIDSController.start_snort_managers(emulation_env_config=emulation_env_config,
                                                    physical_server_ip=physical_server_ip,
                                                    logger=logger)

        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            for ids_image in constants.CONTAINER_IMAGES.SNORT_IDS_IMAGES:
                if ids_image in c.name:
                    status = SnortIDSController.get_snort_idses_monitor_threads_statuses_by_ip_and_port(
                        port=emulation_env_config.snort_ids_manager_config.snort_ids_manager_port,
                        ip=c.docker_gw_bridge_ip)
                    statuses.append(status)
        return statuses

    @staticmethod
    def get_snort_idses_monitor_threads_statuses_by_ip_and_port(
            port: int, ip: str, timeout: int = csle_collector_constants.GRPC.TIMEOUT_SECONDS) -> (
            csle_collector.snort_ids_manager.snort_ids_manager_pb2.SnortIdsMonitorDTO):
        """
        A method that sends a request to the SnortIDSManager with a specific port and ip
        to get the status of the IDS monitor thread

        :param port: the port of the SnortIDSManager
        :param ip: the ip of the SnortIDSManager
        :param timeout: the timeout of the GRPC query
        :return: the status of the SnortIDSManager
        """
        with grpc.insecure_channel(f'{ip}:{port}', options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.snort_ids_manager.snort_ids_manager_pb2_grpc.SnortIdsManagerStub(channel)
            status = \
                csle_collector.snort_ids_manager.query_snort_ids_manager.get_snort_ids_monitor_status(
                    stub=stub, timeout=timeout)
            return status

    @staticmethod
    def get_snort_ids_managers_ips(emulation_env_config: EmulationEnvConfig) -> List[str]:
        """
        A method that extracts the IPs of the snort IDS managers in a given emulation

        :param emulation_env_config: the emulation env config
        :return: the list of IP addresses
        """
        ips = []
        for c in emulation_env_config.containers_config.containers:
            for ids_image in constants.CONTAINER_IMAGES.SNORT_IDS_IMAGES:
                if ids_image in c.name:
                    ips.append(c.docker_gw_bridge_ip)
        return ips

    @staticmethod
    def get_snort_idses_managers_ports(emulation_env_config: EmulationEnvConfig) -> List[int]:
        """
        A method that extracts the ports of the snort IDS managers in a given emulation

        :param emulation_env_config: the emulation env config
        :return: the list of ports
        """
        ports = []
        for c in emulation_env_config.containers_config.containers:
            for ids_image in constants.CONTAINER_IMAGES.SNORT_IDS_IMAGES:
                if ids_image in c.name:
                    ports.append(emulation_env_config.snort_ids_manager_config.snort_ids_manager_port)
        return ports

    @staticmethod
    def get_snort_managers_info(emulation_env_config: EmulationEnvConfig, active_ips: List[str],
                                logger: logging.Logger, physical_server_ip: str) -> SnortIdsManagersInfo:
        """
        Extracts the information of the Snort managers for a given emulation

        :param emulation_env_config: the configuration of the emulation
        :param active_ips: list of active IPs
        :param physical_server_ip: the IP of the physical server
        :param logger: the logger to use for logging
        :return: a DTO with the status of the Snort managers
        """
        snort_ids_managers_ips = SnortIDSController.get_snort_ids_managers_ips(
            emulation_env_config=emulation_env_config)
        snort_ids_managers_ports = \
            SnortIDSController.get_snort_idses_managers_ports(emulation_env_config=emulation_env_config)
        snort_managers_statuses = []
        snort_managers_running = []
        for ip in snort_ids_managers_ips:
            if ip not in active_ips or not EmulationUtil.physical_ip_match(
                    emulation_env_config=emulation_env_config, ip=ip, physical_host_ip=physical_server_ip):
                continue
            running = False
            status = None
            try:
                status = SnortIDSController.get_snort_idses_monitor_threads_statuses_by_ip_and_port(
                    port=emulation_env_config.snort_ids_manager_config.snort_ids_manager_port, ip=ip)
                running = True
            except Exception as e:
                logger.debug(
                    f"Could not fetch Snort IDS manager status on IP:{ip}, error: {str(e)}, {repr(e)}")
            if status is not None:
                snort_managers_statuses.append(status)
            else:
                util = csle_collector.snort_ids_manager.snort_ids_manager_util.SnortIdsManagerUtil
                snort_managers_statuses.append(util.snort_ids_monitor_dto_empty())
            snort_managers_running.append(running)
        execution_id = emulation_env_config.execution_id
        emulation_name = emulation_env_config.name
        snort_manager_info_dto = SnortIdsManagersInfo(
            snort_ids_managers_running=snort_managers_running, ips=snort_ids_managers_ips,
            ports=snort_ids_managers_ports, execution_id=execution_id, emulation_name=emulation_name,
            snort_ids_managers_statuses=snort_managers_statuses)
        return snort_manager_info_dto
